import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.pad import pad_array
from gdsfactory.port import select_ports_electrical
from gdsfactory.routing.get_route_electrical_shortest_path import (
    get_route_electrical_shortest_path,
)


@gf.cell
def add_electrical_pads_top(
    component: Component,
    dy: float = 100.0,
    select_ports=select_ports_electrical,
    **kwargs,
) -> Component:
    """Returns new component with electrical ports connected to top pad array

    Args:
        component:
        dy: vertical spacing
        select_ports: function to select electrical ports
        kwargs:pad settings
            pad: pad element
            pitch: x spacing
            n: number of pads
            **port_settings
    """
    c = Component()
    c.component = component
    ref = c << component
    ports = select_ports(ref.ports)
    ports = list(ports.values())
    pads = c << pad_array(columns=len(ports), orientation=270, **kwargs)
    pads.x = ref.x
    pads.ymin = ref.ymax + dy
    ports_pads = list(pads.ports.values())

    ports_pads = gf.routing.sort_ports.sort_ports_x(ports_pads)
    ports_component = gf.routing.sort_ports.sort_ports_x(ports)

    for p1, p2 in zip(ports_component, ports_pads):
        c.add(get_route_electrical_shortest_path(p1, p2))

    c.add_ports(ref.ports)
    for port in ports:
        c.ports.pop(port.name)
    return c


if __name__ == "__main__":
    # c = demo_mzi()
    # c = demo_straight()
    # c.show()
    import gdsfactory as gf

    c = gf.components.straight_heater_metal()
    c = gf.components.mzi_phase_shifter()
    cc = add_electrical_pads_top(component=c)
    cc.show()



from typing import Optional, Tuple

from numpy import floor

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.via import via1
from gdsfactory.tech import LAYER
from gdsfactory.types import ComponentFactory, Layer


@gf.cell
def via_stack_with_offset(
    layer_via_width_height_offset: Tuple[
        Tuple[Layer, Optional[ComponentFactory], Tuple[float, float], float], ...
    ] = (
        (LAYER.Ppp, None, (10, 10), 0),
        (LAYER.M1, via1, (10, 10), 0),
    ),
    port_orientation: int = 180,
) -> Component:
    """Rectangular transition thru metal layers with offset between layers

    Args:
        layer_via_width_height_offset:
            layer:
            via: factory
            size:
            offset: for next layer
        port_orientation: 180: W0, 0: E0, 90: N0, 270: S0
    """
    c = Component()
    x0 = x1 = y0 = y1 = 0

    i = 0
    for i, (layer, via, size, offset) in enumerate(layer_via_width_height_offset):
        width, height = size
        x0 = -width / 2
        x1 = +width / 2
        y1 = y0 + height
        rect_pts = [(x0, y0), (x1, y0), (x1, y1), (x0, y1)]
        c.add_polygon(rect_pts, layer=layer)

        if via:
            via = via()
            w = via.info["width"]
            h = via.info["height"]
            g = via.info["enclosure"]
            pitch_x = via.info["pitch_x"]
            pitch_y = via.info["pitch_y"]

            nb_vias_x = (width - w - 2 * g) / pitch_x + 1
            nb_vias_y = (height - h - 2 * g) / pitch_y + 1

            nb_vias_x = int(floor(nb_vias_x)) or 1
            nb_vias_y = int(floor(nb_vias_y)) or 1

            cw = (width - (nb_vias_x - 1) * pitch_x - w) / 2
            ch = (height - (nb_vias_y - 1) * pitch_y - h) / 2

            x00 = x0 + cw + w / 2
            y00 = y0 + ch + h / 2 + offset

            for i in range(nb_vias_x):
                for j in range(nb_vias_y):
                    c.add(via.ref(position=(x00 + i * pitch_x, y00 + j * pitch_y)))
        y0 += offset
        y1 = y0 + height
        rect_pts = [(x0, y0), (x1, y0), (x1, y1), (x0, y1)]
        c.add_polygon(rect_pts, layer=layer)

    port_width = height if port_orientation in [0, 180] else width

    if port_orientation not in [0, 90, 270, 180]:
        raise ValueError(
            f"Invalid port_orientation = {port_orientation} not in [0, 90, 180, 270]"
        )
    c.add_port(
        name="e1",
        width=port_width,
        orientation=port_orientation,
        midpoint=(0, y1),
        port_type="electrical",
    )
    return c


if __name__ == "__main__":

    # c = via_stack_with_offset(
    #     layer_via_width_height_offset=(
    #         (LAYER.Ppp, gf.components.via1, (10, 10), 0),
    #         (LAYER.M1, None, (10, 10), 10),
    #     )
    # )
    # c = via_stack_with_offset(
    #     layer_via_width_height_offset=(
    #         (LAYER.Ppp, gf.components.via1, (10, 10), 10),
    #         (LAYER.M1, gf.components.via2, (10, 10), 0),
    #         (LAYER.M2, None, (10, 10), 10),
    #     )
    # )
    # c = via_stack_with_offset(
    #     layer_via_width_height_offset=(
    #         (LAYER.Ppp, gf.components.via1, (5, 10), 0),
    #         (LAYER.M1, gf.components.via2, (5, 10), 10),
    #         (LAYER.M2, gf.components.via3, (5, 10), 0),
    #         # (LAYER.M3, None, 5, 10, 0),
    #     )
    # )
    # c.pprint
    c = via_stack_with_offset()
    # print(c)
    c.show()



from typing import Tuple

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.types import Layer


@cell
def compass(
    size: Tuple[float, float] = (4.0, 2.0),
    layer: Layer = gf.LAYER.WG,
    port_type: str = "electrical",
) -> Component:
    """Rectangular contact pad with centered ports on rectangle edges
    (north, south, east, and west)

    Args:
        size: tuple (4, 2)
        layer: tuple (int, int)

    """
    prefix = "e"

    c = gf.Component()
    r = gf.components.rectangle(size=size, layer=layer)
    dx = size[0]
    dy = size[1]

    rr = r.ref(position=(-dx / 2, -dy / 2))
    c.add(rr)
    c.add_port(
        name=f"{prefix}1",
        midpoint=[-dx / 2, 0],
        width=dy,
        orientation=180,
        layer=layer,
        port_type=port_type,
    )
    c.add_port(
        name=f"{prefix}2",
        midpoint=[0, dy / 2],
        width=dx,
        orientation=90,
        layer=layer,
        port_type=port_type,
    )
    c.add_port(
        name=f"{prefix}3",
        midpoint=[dx / 2, 0],
        width=dy,
        orientation=0,
        layer=layer,
        port_type=port_type,
    )
    c.add_port(
        name=f"{prefix}4",
        midpoint=[0, -dy / 2],
        width=dx,
        orientation=-90,
        layer=layer,
        port_type=port_type,
    )
    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    c = compass(size=(1, 2), layer=gf.LAYER.WG, port_type="optical")
    c.show()



"""Read component GDS, JSON metadata and CSV ports."""
from pathlib import Path
from typing import Union

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.import_gds import import_gds


def gds(gdspath: Union[str, Path], **kwargs) -> Component:
    """Returns Component with ports (CSV) and metadata (JSON) info (if any).
    Args:
        gdspath: path of GDS file
        cellname: cell of the name to import (None) imports top cell
        flatten: if True returns flattened (no hierarchy)
        snap_to_grid_nm: snap to different nm grid (does not snap if False)

    """

    if not gdspath.exists():
        raise FileNotFoundError(f"No such file '{gdspath}'")
    component = import_gds(gdspath)

    gf.read.read_ports(component=component, gdspath=gdspath)
    return component


if __name__ == "__main__":
    from gdsfactory.tests.test_load_component import test_load_component_gds

    test_load_component_gds()



"""add_pin adss a Pin to a port, add_pins adds Pins to all ports:

- pins
- outline

Some functions modify a component without changing its name.
Make sure these functions are inside a new Component or called as a decorator
They without modifying the cell name

"""
import json
from typing import Callable, Optional, Tuple

import numpy as np
from numpy import ndarray

import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.component import Component, ComponentReference
from gdsfactory.port import Port
from gdsfactory.tech import LAYER


def _rotate(v: ndarray, m: ndarray) -> ndarray:
    return np.dot(m, v)


def add_pin_triangle(
    component: Component,
    port: Port,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.TEXT,
) -> None:
    """Add triangle pin with a right angle, pointing out of the port

    Args:
        component:
        port: Port
        layer: for the pin marker
        label_layer: for the label
    """
    p = port

    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dtip = np.array([d, 0])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptip = p.position + _rotate(dtip, rot_mat)
    polygon = [p0, p1, ptip]
    component.add_polygon(polygon, layer=layer)

    if label_layer:
        component.add_label(
            text=str(p.name),
            position=ptip,
            layer=label_layer,
        )


def add_pin_square_inside(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.TEXT,
) -> None:
    """Add square pin towards the inside of the port

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label

    .. code::

           _______________
          |               |
          |               |
          |               |
          ||              |
          ||              |
          |               |
          |      __       |
          |_______________|


    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([-pin_length, -d])
    dtopin = np.array([-pin_length, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)
    if label_layer:
        component.add_label(
            text=str(p.name),
            position=p.midpoint,
            layer=label_layer,
        )


def add_pin_square_double(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.TEXT,
) -> None:
    """Add two square pins: one inside with label, one outside

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label

    .. code::

           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    # outer square
    d = p.width / 2
    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([pin_length / 2, -d])
    dtopin = np.array([pin_length / 2, +d])
    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    # inner square
    d = p.width / 2
    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([-pin_length / 2, -d])
    dtopin = np.array([-pin_length / 2, +d])
    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    x = (p0[0] + ptopin[0]) / 2
    y = (ptopin[1] + pbotin[1]) / 2
    if label_layer:
        component.add_label(
            text=str(p.name),
            position=(x, y),
            layer=label_layer,
        )


def add_pin_square(
    component: Component,
    port: Port,
    pin_length: float = 0.1,
    layer: Tuple[int, int] = LAYER.PORT,
    label_layer: Optional[Tuple[int, int]] = LAYER.PORT,
    port_margin: float = 0.0,
) -> None:
    """Add half out pin to a component.

    Args:
        component:
        port: Port
        pin_length: length of the pin marker for the port
        layer: for the pin marker
        label_layer: for the label
        port_margin: margin to port edge


    .. code::

           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2 + port_margin

    dbot = np.array([pin_length / 2, -d])
    dtop = np.array([pin_length / 2, d])
    dbotin = np.array([-pin_length / 2, -d])
    dtopin = np.array([-pin_length / 2, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    if label_layer:
        component.add_label(
            text=str(p.name),
            position=p.midpoint,
            layer=label_layer,
        )


def add_outline(
    component: Component,
    reference: Optional[ComponentReference] = None,
    layer: Tuple[int, int] = LAYER.DEVREC,
    **kwargs,
) -> None:
    """Adds devices outline bounding box in layer.

    Args:
        component: where to add the markers
        reference: to read outline from
        layer: to add padding
        default: default padding
        top: North padding
        bottom
        right
        left
    """
    c = reference or component
    if hasattr(component, "parent"):
        component = component.parent
    points = get_padding_points(component=c, default=0, **kwargs)
    component.add_polygon(points, layer=layer)


def add_pins(
    component: Component,
    reference: Optional[ComponentReference] = None,
    function: Callable = add_pin_square_inside,
    select_ports: Optional[Callable] = None,
    **kwargs,
) -> None:
    """Add Pin port markers.

    Be careful with this function as it modifies the component

    Args:
        component: to add ports to
        reference:
        function: to add each pin
        select_ports: function to select_ports
        kwargs: add pins function settings

    """
    reference = reference or component
    ports = (
        select_ports(reference.ports).values()
        if select_ports
        else reference.get_ports_list()
    )
    for port in ports:
        function(component=component, port=port, **kwargs)


add_pins_triangle = gf.partial(add_pins, function=add_pin_triangle)


@gf.cell
def add_pins_container(
    component: Component,
    add_pins_function=add_pins_triangle,
) -> Component:
    c = Component()
    ref = c << component
    add_pins_function(component=c, reference=ref)
    c.ref = ref
    return c


def add_settings_label(
    component: Component,
    reference: ComponentReference,
    label_layer: Tuple[int, int] = LAYER.LABEL_SETTINGS,
    ignore: Optional[Tuple[str, ...]] = None,
) -> None:
    """Add settings in label, ignores component.ignore keys.

    Args:
        componnent
        reference
        label_layer:
        ignore: fields to ignore

    """
    settings = reference.get_settings(ignore=ignore)
    settings_string = f"settings={json.dumps(settings, indent=2)}"
    if len(settings_string) > 1024:
        raise ValueError(f"label > 1024 characters: {settings_string}")
    component.add_label(
        position=reference.center, text=settings_string, layer=label_layer
    )


def add_instance_label(
    component: Component,
    reference: ComponentReference,
    instance_name: Optional[str] = None,
    layer: Tuple[int, int] = LAYER.LABEL_INSTANCE,
) -> None:
    """Adds label to a reference in a component."""

    instance_name = (
        instance_name
        or f"{reference.parent.name},{int(reference.x)},{int(reference.y)}"
    )
    x = gf.snap.snap_to_grid(reference.x)
    y = gf.snap.snap_to_grid(reference.y)

    component.add_label(
        text=instance_name,
        position=(x, y),
        layer=layer,
    )


def add_pins_and_outline(
    component: Component,
    reference: ComponentReference,
    add_outline_function: Optional[Callable] = add_outline,
    add_pins_function: Optional[Callable] = add_pins,
    add_settings_function: Optional[Callable] = add_settings_label,
    add_instance_label_function: Optional[Callable] = add_settings_label,
) -> None:
    """Add markers:
    - outline
    - pins for the ports
    - label for the name
    - label for the settings

    Args:
        component: where to add the markers
        reference
        add_outline_function
        add_pins_function: to add pins to ports
        add_settings_function: to add outline around the component
        add_instance_label_function: labels each instance

    """
    if add_outline_function:
        add_outline_function(component=component, reference=reference)
    if add_pins_function:
        add_pins_function(component=component, reference=reference)
    if add_settings_function:
        add_settings_function(component=component, reference=reference)
    if add_instance_label_function:
        add_instance_label_function(component=component, reference=reference)


@gf.cell
def add_pins_to_references(
    component: Component,
    function: Callable = add_pins_triangle,
) -> Component:
    """Add pins to Component references.

    Args:
        component: component to add pins to
        references:
        function: function to add pins
    """
    c = Component()
    c.add_ref(component)

    references = component.references

    for reference in references:
        function(component=c, reference=reference)

    return c


def test_add_pins() -> None:
    c1 = gf.c.straight(length=2)
    c2 = add_pins_container(component=c1)

    p1 = len(c1.get_polygons())
    p2 = len(c2.get_polygons())

    assert p2 == p1 + 2
    assert len(c2.polygons) == 2


if __name__ == "__main__":
    # c = test_add_pins()
    # c.show()

    # c = gf.c.straight(length=2)
    # c = add_pins_container(component=c)

    # c.show(show_ports_suborts=True)

    # c2 = add_pins_container(component=c1)

    # p1 = len(c1.get_polygons())
    # p2 = len(c2.get_polygons())
    # assert p2 == p1 + 2

    # c1 = gf.c.straight_heater_metal(length=2)
    # c2 = add_pins_to_references(c1)
    c = gf.c.ring_single()
    # cc = add_pins_to_references(component=c)
    # cc.show(show_ports=False)
    c.show(show_subports=True)
    c.show(show_ports=True)



"""Dummy filing to keep density constant.
"""
from typing import Iterable, Optional, Tuple, Union

from phidl.geometry import fill_rectangle as _fill_rectangle

from gdsfactory.component import Component
from gdsfactory.component_from import phidl
from gdsfactory.types import Layer


def fill_rectangle(
    component: Component,
    fill_size: Tuple[float, float] = (5.0, 5.0),
    avoid_layers: Union[str, Iterable[Layer]] = "all",
    include_layers: Optional[Iterable[Layer]] = None,
    margin: float = 5.0,
    fill_layers: Iterable[Layer] = (0, 1, 3),
    fill_densities: Iterable[float] = (0.5, 0.25, 0.7),
    fill_inverted: Optional[bool] = None,
    bbox: Optional[Tuple[float, float]] = None,
) -> Component:
    """Creates a rectangular fill pattern and fills all empty areas
    in the input component

    Args:
        Component: Component to be filled
        fill_size: Rectangular size of the fill element
        avoid_layers : 'all' or list of layers
            Layers to be avoided (not filled) in D
        include_layers :
            Layers to be included (filled) in D, supercedes avoid_layers
        margin :
            Margin spacing around avoided areas -- fill will not come within
            `margin` of the geometry in D
        fill_layers : list of layers. fill pattern layers
        fill_densities : float between 0 and 1
            Defines the fill pattern density (1.0 == fully filled)
        fill_inverted : Inverts the fill pattern
        bbox: array-like[2][2]
            Limit the fill pattern to the area defined by this bounding box

    """

    component_filled = _fill_rectangle(
        component,
        fill_size=fill_size,
        avoid_layers=avoid_layers,
        include_layers=include_layers,
        margin=margin,
        fill_layers=fill_layers,
        fill_densities=fill_densities,
        fill_inverted=fill_inverted,
        bbox=bbox,
    )
    return phidl(component_filled)


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight()
    c = gf.add_padding(c)
    c << fill_rectangle(
        c,
        fill_layers=((2, 0),),
        fill_densities=(1.0),
        avoid_layers=((1, 0),),
        # bbox=(100.0, 100.0),
    )
    c.show()



"""Returns Component from YAML syntax.

instances:
    mzi:
        component: mzi_phase_shifter
        settings:
            length_x: 50

    pads:
        component: pad_array
        settings:
            n: 2
            port_names:
                - e4

placements:
    mzi:
        x: 0
    pads:
        y: 200
        x: mzi,cc
ports:
    o1: mzi,o1
    o2: mzi,o2


routes:
    electrical:
        links:
            mzi,etop_e1: pads,e4_0
            mzi,etop_e2: pads,e4_1

        settings:
            layer: [31, 0]
            width: 10
            radius: 10

"""

import functools
import io
import pathlib
import warnings
from typing import IO, Any, Callable, Dict, List, Optional, Union

import numpy as np
import omegaconf
from omegaconf import OmegaConf

from gdsfactory.add_pins import add_instance_label
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components import factory
from gdsfactory.cross_section import cross_section_factory
from gdsfactory.routing.factories import routing_strategy as routing_strategy_factories
from gdsfactory.types import ComponentFactoryDict, CrossSectionFactory, Route

valid_placement_keys = ["x", "y", "dx", "dy", "rotation", "mirror", "port"]


valid_top_level_keys = [
    "name",
    "instances",
    "placements",
    "connections",
    "ports",
    "routes",
]

valid_anchor_point_keywords = [
    "ce",
    "cw",
    "nc",
    "ne",
    "nw",
    "sc",
    "se",
    "sw",
    "center",
    "cc",
]
# refer to an (x,y) Point

valid_anchor_value_keywords = [
    "south",
    "west",
    "east",
    "north",
]
# refer to a singular (x or y) value

valid_anchor_keywords = valid_anchor_point_keywords + valid_anchor_value_keywords
# full set of valid anchor keywords (either referring to points or values)

valid_route_keys = [
    "links",
    "settings",
    "routing_strategy",
]
# Recognized keys within a YAML route definition


def _get_anchor_point_from_name(
    ref: ComponentReference, anchor_name: str
) -> Optional[np.ndarray]:
    if anchor_name in valid_anchor_point_keywords:
        pt = getattr(ref.size_info, anchor_name)
        return pt
    elif anchor_name in ref.ports:
        return ref.ports[anchor_name].position
    else:
        return None


def _get_anchor_value_from_name(
    ref: ComponentReference, anchor_name: str, return_value: str
) -> Optional[float]:
    if anchor_name in valid_anchor_value_keywords:
        v = getattr(ref.size_info, anchor_name)
        return v
    else:
        anchor_point = _get_anchor_point_from_name(ref, anchor_name)
        if anchor_point is None:
            return None
        if return_value == "x":
            return anchor_point[0]
        elif return_value == "y":
            return anchor_point[1]
        else:
            raise ValueError("Expected x or y as return_value.")


def place(
    placements_conf: Dict[str, Dict[str, Union[int, float, str]]],
    connections_by_transformed_inst: Dict[str, Dict[str, str]],
    instances: Dict[str, ComponentReference],
    encountered_insts: List[str],
    instance_name: Optional[str] = None,
    all_remaining_insts: Optional[List[str]] = None,
) -> None:
    """Place instance_name with placements_conf config.

    Args:
        placements_conf: Dict of instance_name to placement (x, y, rotation ...)
        connections_by_transformed_inst: Dict of connection attributes.
            keyed by the name of the instance which should be transformed
        instances: Dict of references
        encountered_insts: list of encountered_instances
        instance_name: instance_name to place
        all_remaining_insts: list of all the remaining instances to place
            instances pop from this instrance as they are placed
    """
    if not all_remaining_insts:
        return
    if instance_name is None:
        instance_name = all_remaining_insts.pop(0)
    else:
        all_remaining_insts.remove(instance_name)

    if instance_name in encountered_insts:
        encountered_insts.append(instance_name)
        loop_str = " -> ".join(encountered_insts)
        raise ValueError(
            f"circular reference in placement for {instance_name}! Loop: {loop_str}"
        )
    encountered_insts.append(instance_name)
    ref = instances[instance_name]

    if instance_name in placements_conf:
        placement_settings = placements_conf[instance_name] or {}
        if not isinstance(placement_settings, omegaconf.DictConfig):
            raise ValueError(
                f"Invalid placement {placement_settings} from {valid_placement_keys}"
            )
        for k in placement_settings.keys():
            if k not in valid_placement_keys:
                raise ValueError(f"Invalid placement {k} from {valid_placement_keys}")

        x = placement_settings.get("x")
        y = placement_settings.get("y")
        dx = placement_settings.get("dx")
        dy = placement_settings.get("dy")
        port = placement_settings.get("port")
        rotation = placement_settings.get("rotation")
        mirror = placement_settings.get("mirror")

        if port:
            a = _get_anchor_point_from_name(ref, port)
            if a is None:
                raise ValueError(
                    f"Port {port} is neither a valid port on {ref.parent.name}"
                    " nor a recognized anchor keyword.\n"
                    "Valid ports: \n"
                    f"{list(ref.ports.keys())}. \n"
                    "Valid keywords: \n"
                    f"{valid_anchor_point_keywords}",
                )
            ref.x -= a[0]
            ref.y -= a[1]
        if x:
            if isinstance(x, str):
                if not len(x.split(",")) == 2:
                    raise ValueError(
                        f"You can define x as `x: instaceName,portName` got `x: {x}`"
                    )
                instance_name_ref, port_name = x.split(",")
                if instance_name_ref in all_remaining_insts:
                    place(
                        placements_conf,
                        connections_by_transformed_inst,
                        instances,
                        encountered_insts,
                        instance_name_ref,
                        all_remaining_insts,
                    )
                if instance_name_ref not in instances:
                    raise ValueError(
                        f"instaceName `{instance_name_ref}` not in {list(instances.keys())}, "
                        f"you can define x as `x: instaceName,portName`, got `x: {x}`"
                    )
                if (
                    port_name not in instances[instance_name_ref].ports
                    and port_name not in valid_anchor_keywords
                ):
                    raise ValueError(
                        f"port = `{port_name}` not in {list(instances[instance_name_ref].ports.keys())}"
                        f" or in valid anchors {valid_anchor_keywords} for {instance_name_ref}, "
                        f"you can define x as `x: instaceName,portName`, got `x: {x}`"
                    )

                x = _get_anchor_value_from_name(
                    instances[instance_name_ref], port_name, "x"
                )
            ref.x += x
        if y:
            if isinstance(y, str):
                if not len(y.split(",")) == 2:
                    raise ValueError(
                        f"You can define y as `y: instaceName,portName` got `y: {y}`"
                    )
                instance_name_ref, port_name = y.split(",")
                if instance_name_ref in all_remaining_insts:
                    place(
                        placements_conf,
                        connections_by_transformed_inst,
                        instances,
                        encountered_insts,
                        instance_name_ref,
                        all_remaining_insts,
                    )
                if instance_name_ref not in instances:
                    raise ValueError(
                        f"{instance_name_ref} not in {list(instances.keys())}, "
                        f"you can define y as `y: instaceName,portName`, got `y: {y}`"
                    )
                if (
                    port_name not in instances[instance_name_ref].ports
                    and port_name not in valid_anchor_keywords
                ):
                    raise ValueError(
                        f"port = {port_name} not in {list(instances[instance_name_ref].ports.keys())} "
                        f"or in valid anchors {valid_anchor_keywords} for {instance_name_ref}, "
                        f"you can define y as `y: instaceName,portName`, got `y: {y}`"
                    )

                y = _get_anchor_value_from_name(
                    instances[instance_name_ref], port_name, "y"
                )
            ref.y += y
        if dx:
            ref.x += dx
        if dy:
            ref.y += dy
        if mirror:
            if mirror is True and port:
                ref.reflect_h(x0=_get_anchor_value_from_name(ref, port, "x"))
            elif mirror is True:
                if x:
                    ref.reflect_h(x0=x)
                else:
                    ref.reflect_h()
            elif mirror is False:
                pass
            elif isinstance(mirror, str):
                ref.reflect_h(port_name=mirror)
            elif isinstance(mirror, (int, float)):
                ref.reflect_h(x0=mirror)
            else:
                raise ValueError(
                    f"{mirror} can only be a port name {ref.ports.keys()}"
                    ", a x value or True/False"
                )

        if rotation:
            if port:
                ref.rotate(rotation, center=_get_anchor_point_from_name(ref, port))
            else:
                x, y = ref.origin
                ref.rotate(rotation, center=(x, y))
                # ref.rotate(rotation, center=(ref.x, ref.y))

    if instance_name in connections_by_transformed_inst:
        conn_info = connections_by_transformed_inst[instance_name]
        instance_dst_name = conn_info["instance_dst_name"]
        if instance_dst_name in all_remaining_insts:
            place(
                placements_conf,
                connections_by_transformed_inst,
                instances,
                encountered_insts,
                instance_dst_name,
                all_remaining_insts,
            )

        make_connection(instances=instances, **conn_info)
        # placements_conf.pop(instance_name)


def transform_connections_dict(connections_conf: Dict[str, str]) -> Dict[str, Dict]:
    """Returns Dict with source_instance_name key and connection properties."""
    if not connections_conf:
        return {}
    attrs_by_src_inst = {}
    for port_src_string, port_dst_string in connections_conf.items():
        instance_src_name, port_src_name = port_src_string.split(",")
        instance_dst_name, port_dst_name = port_dst_string.split(",")
        attrs_by_src_inst[instance_src_name] = {
            "instance_src_name": instance_src_name,
            "port_src_name": port_src_name,
            "instance_dst_name": instance_dst_name,
            "port_dst_name": port_dst_name,
        }
    return attrs_by_src_inst


def make_connection(
    instance_src_name: str,
    port_src_name: str,
    instance_dst_name: str,
    port_dst_name: str,
    instances: Dict[str, ComponentReference],
) -> None:
    instance_src_name = instance_src_name.strip()
    instance_dst_name = instance_dst_name.strip()
    port_src_name = port_src_name.strip()
    port_dst_name = port_dst_name.strip()

    assert (
        instance_src_name in instances
    ), f"{instance_src_name} not in {list(instances.keys())}"
    assert (
        instance_dst_name in instances
    ), f"{instance_dst_name} not in {list(instances.keys())}"
    instance_src = instances[instance_src_name]
    instance_dst = instances[instance_dst_name]

    assert port_src_name in instance_src.ports, (
        f"{port_src_name} not in {list(instance_src.ports.keys())} for"
        f" {instance_src_name} "
    )
    assert port_dst_name in instance_dst.ports, (
        f"{port_dst_name} not in {list(instance_dst.ports.keys())} for"
        f" {instance_dst_name}"
    )
    port_dst = instance_dst.ports[port_dst_name]
    instance_src.connect(port=port_src_name, destination=port_dst)


sample_mmis = """
name:
    mmis

instances:
    mmi_long:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 10
    mmi_short:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 5

placements:
    mmi_long:
        rotation: 180
        x: 100
        y: 100

routes:
    route_name1:
        links:
            mmi_short,o2: mmi_long,o1

ports:
    o1: mmi_short,o1
    o2: mmi_long,o2
    o3: mmi_long,o3
"""


def component_from_yaml(
    yaml_str: Union[str, pathlib.Path, IO[Any]],
    component_factory: ComponentFactoryDict = factory,
    routing_strategy: Dict[str, Callable] = routing_strategy_factories,
    cross_section_factory: Dict[str, CrossSectionFactory] = cross_section_factory,
    label_instance_function: Callable = add_instance_label,
    **kwargs,
) -> Component:
    """Returns a Component defined in YAML file or string.

    Args:
        yaml: YAML IO describing Component file or string (with newlines)
            (instances, placements, routes, ports, connections, names)
        component_factory: dict of functions {factory_name: factory_function}
        routing_strategy: for links
        label_instance_function: to label each instance
        kwargs: cache, pins ... to pass to all factories

    Returns:
        Component

    .. code::

        valid properties:
        name: name of Component
        instances:
            name:
                component:
                settings (Optional)
        placements:
            x: Optional[float, str]  str can be instanceName,portName
            y: Optional[float, str]
            rotation: Optional[float]
            mirror: Optional[bool, float] float is x mirror axis
            port: Optional[str] port anchor
        connections (Optional): between instances
        ports (Optional): defines ports to expose
        routes (Optional): defines bundles of routes
            routeName:
            library: optical
            links:
                instance1,port1: instance2,port2


    .. code::

        instances:
            mmi_bot:
              component: mmi1x2
              settings:
                width_mmi: 4.5
                length_mmi: 10
            mmi_top:
              component: mmi1x2
              settings:
                width_mmi: 4.5
                length_mmi: 5

        placements:
            mmi_top:
                port: o1
                x: 0
                y: 0
            mmi_bot:
                port: o1
                x: mmi_top,o2
                y: mmi_top,o2
                dx: 30
                dy: -30
        routes:
            optical:
                library: optical
                links:
                    mmi_top,o3: mmi_bot,o1

    """
    yaml_str = (
        io.StringIO(yaml_str)
        if isinstance(yaml_str, str) and "\n" in yaml_str
        else yaml_str
    )

    conf = OmegaConf.load(yaml_str)  # nicer loader than conf = yaml.safe_load(yaml_str)
    for key in conf.keys():
        assert key in valid_top_level_keys, f"{key} not in {list(valid_top_level_keys)}"

    instances = {}
    routes = {}
    name = conf.get("name", "Unnamed")
    c = Component(name)
    placements_conf = conf.get("placements")
    routes_conf = conf.get("routes")
    ports_conf = conf.get("ports")
    connections_conf = conf.get("connections")
    instances_dict = conf["instances"]

    for instance_name in instances_dict:
        instance_conf = instances_dict[instance_name]
        component_type = instance_conf["component"]
        assert (
            component_type in component_factory
        ), f"{component_type} not in {list(component_factory.keys())}"

        settings = instance_conf.get("settings", {})
        settings = OmegaConf.to_container(settings, resolve=True) if settings else {}
        settings.update(**kwargs)

        if "cross_section" in settings:
            name_or_dict = settings["cross_section"]
            if isinstance(name_or_dict, str):
                cross_section = cross_section_factory[name_or_dict]
            elif isinstance(name_or_dict, dict):
                name = name_or_dict.pop("function")
                cross_section = functools.partial(
                    cross_section_factory[name], **name_or_dict
                )
            else:
                raise ValueError(f"invalid type for cross_section={name_or_dict}")
            settings["cross_section"] = cross_section

        ci = component_factory[component_type](**settings)
        ref = c << ci
        instances[instance_name] = ref

    placements_conf = dict() if placements_conf is None else placements_conf

    connections_by_transformed_inst = transform_connections_dict(connections_conf)
    components_to_place = set(placements_conf.keys())
    components_with_placement_conflicts = components_to_place.intersection(
        connections_by_transformed_inst.keys()
    )
    for instance_name in components_with_placement_conflicts:
        placement_settings = placements_conf[instance_name]
        if "x" in placement_settings or "y" in placement_settings:
            warnings.warn(
                f"YAML defined: ({', '.join(components_with_placement_conflicts)}) "
                + "with both connection and placement. Please use one or the other.",
            )

    all_remaining_insts = list(
        set(placements_conf.keys()).union(set(connections_by_transformed_inst.keys()))
    )

    while all_remaining_insts:
        place(
            placements_conf=placements_conf,
            connections_by_transformed_inst=connections_by_transformed_inst,
            instances=instances,
            encountered_insts=list(),
            all_remaining_insts=all_remaining_insts,
        )

    for instance_name in instances_dict:
        label_instance_function(
            component=c, instance_name=instance_name, reference=instances[instance_name]
        )

    if routes_conf:
        for route_alias in routes_conf:
            route_names = []
            ports1 = []
            ports2 = []
            routes_dict = routes_conf[route_alias]
            for key in routes_dict.keys():
                if key not in valid_route_keys:
                    raise ValueError(
                        f"`{route_alias}` key=`{key}` not in {valid_route_keys}"
                    )

            settings = routes_dict.pop("settings", {})
            settings = (
                OmegaConf.to_container(settings, resolve=True) if settings else {}
            )
            if "cross_section" in settings:
                name_or_dict = settings["cross_section"]
                if isinstance(name_or_dict, str):
                    cross_section = cross_section_factory[name_or_dict]
                elif isinstance(name_or_dict, dict):
                    name = name_or_dict.pop("function")
                    cross_section = functools.partial(
                        cross_section_factory[name], **name_or_dict
                    )
                else:
                    raise ValueError(f"invalid type for cross_section={name_or_dict}")
                settings["cross_section"] = cross_section
            routing_strategy_name = routes_dict.pop("routing_strategy", "get_bundle")
            if routing_strategy_name not in routing_strategy:
                raise ValueError(
                    f"function `{routing_strategy_name}` not in routing_strategy {list(routing_strategy.keys())}"
                )

            if "links" not in routes_dict:
                raise ValueError(
                    f"You need to define links for the `{route_alias}` route"
                )
            links_dict = routes_dict["links"]

            for port_src_string, port_dst_string in links_dict.items():

                if ":" in port_src_string:
                    src, src0, src1 = [s.strip() for s in port_src_string.split(":")]
                    dst, dst0, dst1 = [s.strip() for s in port_dst_string.split(":")]
                    instance_src_name, port_src_name = [
                        s.strip() for s in src.split(",")
                    ]
                    instance_dst_name, port_dst_name = [
                        s.strip() for s in dst.split(",")
                    ]

                    src0 = int(src0)
                    src1 = int(src1)
                    dst0 = int(dst0)
                    dst1 = int(dst1)

                    if src1 > src0:
                        ports1names = [
                            f"{port_src_name}{i}" for i in range(src0, src1 + 1, 1)
                        ]
                    else:
                        ports1names = [
                            f"{port_src_name}{i}" for i in range(src0, src1 - 1, -1)
                        ]

                    if dst1 > dst0:
                        ports2names = [
                            f"{port_dst_name}{i}" for i in range(dst0, dst1 + 1, 1)
                        ]
                    else:
                        ports2names = [
                            f"{port_dst_name}{i}" for i in range(dst0, dst1 - 1, -1)
                        ]

                    assert len(ports1names) == len(ports2names)
                    route_names += [
                        f"{instance_src_name},{i}:{instance_dst_name},{j}"
                        for i, j in zip(ports1names, ports2names)
                    ]

                    instance_src = instances[instance_src_name]
                    instance_dst = instances[instance_dst_name]

                    for port_src_name in ports1names:
                        assert port_src_name in instance_src.ports, (
                            f"{port_src_name} not in {list(instance_src.ports.keys())}"
                            f"for {instance_src_name} "
                        )
                        ports1.append(instance_src.ports[port_src_name])

                    for port_dst_name in ports2names:
                        assert port_dst_name in instance_dst.ports, (
                            f"{port_dst_name} not in {list(instance_dst.ports.keys())}"
                            f"for {instance_dst_name}"
                        )
                        ports2.append(instance_dst.ports[port_dst_name])

                    # print(ports1)
                    # print(ports2)
                    # print(route_names)

                else:
                    instance_src_name, port_src_name = port_src_string.split(",")
                    instance_dst_name, port_dst_name = port_dst_string.split(",")

                    instance_src_name = instance_src_name.strip()
                    instance_dst_name = instance_dst_name.strip()
                    port_src_name = port_src_name.strip()
                    port_dst_name = port_dst_name.strip()

                    assert (
                        instance_src_name in instances
                    ), f"{instance_src_name} not in {list(instances.keys())}"
                    assert (
                        instance_dst_name in instances
                    ), f"{instance_dst_name} not in {list(instances.keys())}"

                    instance_src = instances[instance_src_name]
                    instance_dst = instances[instance_dst_name]

                    assert port_src_name in instance_src.ports, (
                        f"{port_src_name} not in {list(instance_src.ports.keys())} for"
                        f" {instance_src_name} "
                    )
                    assert port_dst_name in instance_dst.ports, (
                        f"{port_dst_name} not in {list(instance_dst.ports.keys())} for"
                        f" {instance_dst_name}"
                    )

                    ports1.append(instance_src.ports[port_src_name])
                    ports2.append(instance_dst.ports[port_dst_name])
                    route_name = f"{port_src_string}:{port_dst_string}"
                    route_names.append(route_name)

            routing_function = routing_strategy[routing_strategy_name]
            route_or_route_list = routing_function(
                ports1=ports1,
                ports2=ports2,
                **settings,
            )

            # FIXME, be more consistent
            if isinstance(route_or_route_list, list):
                for route_name, route_dict in zip(route_names, route_or_route_list):
                    c.add(route_dict.references)
                    routes[route_name] = route_dict.length
            elif isinstance(route_or_route_list, Route):
                c.add(route_or_route_list.references)
                routes[route_name] = route_or_route_list.length
            else:
                raise ValueError(f"{route_or_route_list} needs to be a Route or a list")

    if ports_conf:
        assert hasattr(ports_conf, "items"), f"{ports_conf} needs to be a dict"
        for port_name, instance_comma_port in ports_conf.items():
            instance_name, instance_port_name = instance_comma_port.split(",")
            instance_name = instance_name.strip()
            instance_port_name = instance_port_name.strip()
            assert (
                instance_name in instances
            ), f"{instance_name} not in {list(instances.keys())}"
            instance = instances[instance_name]
            assert instance_port_name in instance.ports, (
                f"{instance_port_name} not in {list(instance.ports.keys())} for"
                f" {instance_name} "
            )
            c.add_port(port_name, port=instance.ports[instance_port_name])
    c.routes = routes
    c.instances = instances
    return c


if __name__ == "__main__":
    for k in factory.keys():
        print(k)
    # c = component_from_yaml(sample_mmis)
    # print(c.get_settings()["info"])

    # from gdsfactory.tests.test_component_from_yaml import yaml_anchor

    # c = component_from_yaml(yaml_anchor)
    # c.show()

    # c = test_connections_regex()
    # c = component_from_yaml(sample_regex_connections)
    # c = component_from_yaml(sample_regex_connections_backwards)
    # c = test_docstring_sample()
    # c = test_connections()
    # c = component_from_yaml(sample_mirror_simple)
    # c = test_connections_2x2()
    # c = test_connections_different_factory()
    # test_connections_different_link_factory()
    # test_connections_waypoints()
    # test_mirror()
    # c = component_from_yaml(sample_different_link_factory)
    # c = test_mirror()
    # c = component_from_yaml(sample_waypoints)
    # c = component_from_yaml(sample_2x2_connections)
    # c = component_from_yaml(sample_mmis)

    # c = component_from_yaml(sample_connections)
    # assert len(c.get_dependencies()) == 3
    # test_component_from_yaml()
    # test_component_from_yaml_with_routing()
    # print(c.ports)
    # c = gf.routing.add_fiber_array(c)



"""Combine multiple JSONs into one."""

import dataclasses
import json
from pathlib import Path
from typing import Any, Dict, Iterable, Optional

from omegaconf.dictconfig import DictConfig

from gdsfactory.config import CONFIG, TECH, logger, write_config


def merge_json(
    doe_directory: Path = CONFIG["doe_directory"],
    gds_directory: Path = CONFIG["gds_directory"],
    extra_directories: Optional[Iterable[Path]] = None,
    jsonpath: Path = CONFIG["mask_directory"] / "metadata.json",
    json_version: int = 6,
    config: DictConfig = TECH,
) -> Dict[str, Any]:
    """Combine several JSON files from config.yml
    in the root of the mask directory, gets mask_name from there

    Args:
        doe_directory: defaults to current working directory
        extra_directories: list of extra_directories
        jsonpath
        json_version:
        config

    """
    logger.debug("Merging JSON files:")
    cells = {}
    extra_directories = extra_directories or []
    config = dataclasses.asdict(config)
    config.pop("library", "")

    for directory in extra_directories + [doe_directory]:
        for filename in directory.glob("*/*.json"):
            logger.debug(filename)
            with open(filename, "r") as f:
                data = json.load(f)
                cells.update(data.get("cells"))

    does = {d.stem: json.loads(open(d).read()) for d in doe_directory.glob("*.json")}
    metadata = dict(
        json_version=json_version,
        cells=cells,
        does=does,
        config=config,
    )

    write_config(metadata, jsonpath)
    logger.info(f"Wrote  metadata in {jsonpath}")
    return metadata


if __name__ == "__main__":
    from pprint import pprint

    d = merge_json()
    pprint(d)

    # print(config["module_versions"])
    # pprint(d['does'])

    # with open(jsonpath, "w") as f:
    #     f.write(json.dumps(d, indent=2))



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler180
from gdsfactory.components.component_sequence import component_sequence
from gdsfactory.components.taper_from_csv import taper_0p5_to_3_l36
from gdsfactory.types import ComponentFactory


@gf.cell
def cutback_component(
    component: ComponentFactory = taper_0p5_to_3_l36,
    cols: int = 4,
    rows: int = 5,
    bend_radius: int = 10,
    port1: str = "o1",
    port2: str = "o2",
    middle_couples: int = 2,
) -> Component:
    """Flips the component, good for tapers that end in wide straights

    Args:
        component
        cols
        rows
        bend_radius: for bend

    """
    component = component() if callable(component) else component
    bend180 = bend_euler180(radius=bend_radius)

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component = {
        "A": (component, port1, port2),
        "B": (component, port2, port1),
        "D": (bend180, "o1", "o2"),
        "C": (bend180, "o2", "o1"),
    }

    # Generate the sequence of staircases

    s = ""
    for i in range(rows):
        s += "AB" * cols
        s += "D" if i % 2 == 0 else "C"

    s = s[:-1]
    s += "AB" * middle_couples

    for i in range(rows):
        s += "AB" * cols
        s += "D" if (i + rows) % 2 == 0 else "C"

    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(sequence=s, symbol_to_component=symbol_to_component)
    c.update_settings(n_devices=len(s))
    return c


@gf.cell
def cutback_component_flipped(
    component: ComponentFactory = taper_0p5_to_3_l36,
    cols: int = 4,
    rows: int = 5,
    bend_radius: int = 10,
    port1: str = "o2",
    port2: str = "o1",
    middle_couples: int = 2,
) -> Component:
    component = component() if callable(component) else component
    bend180 = bend_euler180(radius=bend_radius)

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component = {
        "A": (component, port1, port2),
        "B": (component, port2, port1),
        "D": (bend180, "o1", "o2"),
        "C": (bend180, "o2", "o1"),
    }

    # Generate the sequence of staircases

    s = ""
    for i in range(rows):
        s += "AB" * cols
        s += "C" if i % 2 == 0 else "D"

    s = s[:-1]
    s += "AB" * middle_couples

    for i in range(rows):
        s += "AB" * cols
        s += "D" if (i + rows + 1) % 2 == 0 else "C"

    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(sequence=s, symbol_to_component=symbol_to_component)
    c.update_settings(n_devices=len(s))
    return c


if __name__ == "__main__":
    c = cutback_component()
    c = cutback_component_flipped()
    c.show()



"""
implement basic cell cache

"""

import pathlib

import pydantic

from gdsfactory.component import Component
from gdsfactory.config import home_path, logger
from gdsfactory.import_gds import import_gds
from gdsfactory.name import clean_name, clean_value, get_name_short
from gdsfactory.types import ComponentFactory, PathType

cwd = pathlib.Path.cwd()


@pydantic.validate_arguments
def filecache(
    component_function: ComponentFactory,
    dirpath: PathType = home_path / "filepath",
    overwrite: bool = False,
    flatten: bool = True,
    with_metadata: bool = False,
) -> Component:
    """implements a basic file cache.
    Only builds component function not found in file cache.
    Flattens cells to ensure no naming conflicts
    """
    name = get_name_short(clean_name(clean_value(component_function)))
    gdspath = dirpath / f"{name}.gds"

    if gdspath.exists() and not overwrite:
        logger.info(f"loading filecache GDS {gdspath}")
        component = import_gds(gdspath)
    else:
        logger.info(f"writing filecache GDS {gdspath}")
        component = component_function()
        if flatten:
            component.flatten()

        if with_metadata:
            component.write_gds_with_metadata(gdspath=gdspath)
        else:
            component.write_gds(gdspath=gdspath)
    return component


if __name__ == "__main__":
    import gdsfactory as gf

    # c = filecache(gf.c.mzi)
    # c = gf.c.mzi()
    c = filecache(gf.c.straight_heater_metal_90_90)
    c.show()



""" You can import gdsfactory.as gf

functions:
    - import_gds(): returns a Component from a GDS

classes:

    - Component
    - Port
    - CONFIG
    - TECH

modules:

    - c: components
    - routing

isort:skip_file
"""

from functools import partial
from phidl import quickplot as plot
import phidl.geometry as pg
from phidl.device_layout import Group, Path

# NOTE: import order matters. Only change the order if you know what you are doing
from gdsfactory.filecache import filecache
from gdsfactory.component import Component, ComponentReference, copy
from gdsfactory.config import CONFIG, call_if_func
from gdsfactory.port import Port
from gdsfactory.port import port_array
from gdsfactory.cell import cell
from gdsfactory.cell import cell_without_validator
from gdsfactory.cell import clear_cache
from gdsfactory.tech import LAYER
from gdsfactory.show import show
from gdsfactory.write_doe import write_doe
from gdsfactory.import_gds import import_gds
from gdsfactory.cross_section import CrossSection

from gdsfactory import cross_section
from gdsfactory import asserts
from gdsfactory import components
from gdsfactory import routing
from gdsfactory import bias
from gdsfactory import klive
from gdsfactory import port
from gdsfactory import types
from gdsfactory import path
from gdsfactory import snap
from gdsfactory import tech
from gdsfactory import containers
from gdsfactory import component_from
from gdsfactory import read
from gdsfactory import remove
from gdsfactory import layers
from gdsfactory import add_termination


from gdsfactory.tech import TECH
from gdsfactory.component_from_yaml import component_from_yaml
from gdsfactory.add_tapers import add_tapers
from gdsfactory.add_padding import (
    add_padding,
    get_padding_points,
    add_padding_container,
)
from gdsfactory.add_pins import add_pins, add_pins_to_references
from gdsfactory.fill import fill_rectangle
from gdsfactory.pack import pack
from gdsfactory.grid import grid
from gdsfactory.offset import offset
from gdsfactory.boolean import boolean
from gdsfactory.rotate import rotate


c = components
lys = layers.load_lyp_generic()

__all__ = [
    "CONFIG",
    "Component",
    "ComponentReference",
    "CrossSection",
    "Group",
    "LAYER",
    "Path",
    "Port",
    "TECH",
    "add_padding",
    "add_padding_container",
    "add_pins",
    "add_pins_to_references",
    "add_termination",
    "add_tapers",
    "asserts",
    "bias",
    "boolean",
    "c",
    "call_if_func",
    "cell",
    "cell_without_validator",
    "clear_cache",
    "component_from",
    "component_from_yaml",
    "components",
    "containers",
    "copy",
    "cross_section",
    "fill_rectangle",
    "filecache",
    "get_padding_points",
    "grid",
    "import_gds",
    "klive",
    "layers",
    "offset",
    "pack",
    "path",
    "partial",
    "pg",
    "plot",
    "port",
    "port_array",
    "read",
    "remove",
    "rotate",
    "routing",
    "show",
    "snap",
    "tech",
    "types",
    "write_doe",
]
__version__ = "3.1.9"


if __name__ == "__main__":
    print(__all__)



from pytest_regressions.data_regression import DataRegressionFixture
from pytest_regressions.num_regression import NumericRegressionFixture

import gdsfactory as gf


def test_load_component_gds() -> None:
    gdspath = gf.CONFIG["gdsdir"] / "straight.gds"
    c = gf.component_from.gds(gdspath)
    assert c.hash_geometry() == "4b8f6646dcf60b78b905ac0c1665a35f119be32a"


def test_load_component_ports(num_regression: NumericRegressionFixture) -> None:
    gdspath = gf.CONFIG["gdsdir"] / "straight.gds"
    c = gf.component_from.gds(gdspath)
    num_regression.check(c.get_ports_array())


def test_load_component_settings(data_regression: DataRegressionFixture) -> None:
    gdspath = gf.CONFIG["gdsdir"] / "straight.gds"
    c = gf.component_from.gds(gdspath)
    data_regression.check(c.get_settings())


if __name__ == "__main__":
    test_load_component_gds()



import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.tech import LAYER
from gdsfactory.types import Layer


@gf.cell
def verniers(
    width_min: float = 0.1,
    width_max: float = 0.5,
    gap: float = 0.1,
    size_max: int = 11,
    layer_label: Layer = LAYER.LABEL,
    **kwargs
) -> Component:
    c = gf.Component()
    y = 0

    widths = np.linspace(width_min, width_max, int(size_max / (width_max + gap)))

    for width in widths:
        w = c << gf.components.straight(width=width, length=size_max, **kwargs)
        y += width / 2
        w.y = y
        c.add_label(text=str(int(width * 1e3)), position=(0, y), layer=layer_label)
        y += width / 2 + gap

    return c


if __name__ == "__main__":
    c = verniers()
    c.show()



import pytest
from pytest_regressions.data_regression import DataRegressionFixture
from pytest_regressions.num_regression import NumericRegressionFixture

from gdsfactory.component import Component
from gdsfactory.components import factory
from gdsfactory.difftest import difftest

skip_test = {
    "version_stamp",
    "extend_ports_list",
    "extend_port",
    "component_sequence",
    "mzi_arm",
}

components_to_test = set(factory.keys()) - skip_test


@pytest.fixture(params=components_to_test, scope="function")
def component(request) -> Component:
    return factory[request.param]()


def test_gds(component: Component) -> None:
    """Avoid regressions in GDS geometry shapes and layers."""
    difftest(component)


def test_settings(component: Component, data_regression: DataRegressionFixture) -> None:
    """Avoid regressions when exporting settings."""
    data_regression.check(component.get_settings())


def test_ports(component: Component, num_regression: NumericRegressionFixture) -> None:
    """Avoid regressions in port names and locations."""
    if component.ports:
        num_regression.check(component.get_ports_array())


def test_assert_ports_on_grid(component: Component):
    """Ensure ports are on grid."""
    component.assert_ports_on_grid()


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.coupler(length=1.0 + 1e-4)
    # c = gf.components.coupler()
    # c.assert_ports_on_grid()
    print(c.ports)



import gdspy

from gdsfactory.types import PathType


def check_duplicate_cells(gdspath: PathType):
    """
    FIXME at gdspy level
    """
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(gdspath)
    # component = gf.import_gds(gdspath)
    # cells = component.get_dependencies()
    # cell_names = [cell.name for cell in list(cells)]
    # cell_names_unique = set(cell_names)

    # if len(cell_names) != len(set(cell_names)):
    #     for cell_name in cell_names_unique:
    #         cell_names.remove(cell_name)

    #     cell_names_duplicated = "\n".join(set(cell_names))
    #     raise ValueError(
    #         f"Duplicated cell names in {component.name}:\n{cell_names_duplicated}"
    #     )


if __name__ == "__main__":
    # check_duplicate_cells("gds.gds")

    gdspath = "gds.gds"
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(gdspath)



""" bends with grating couplers inside the spiral
"""
from typing import Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler, bend_euler180
from gdsfactory.components.straight import straight
from gdsfactory.cross_section import strip
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import ComponentFactory, CrossSectionFactory


def get_bend_port_distances(bend: Component) -> Tuple[float, float]:
    p0, p1 = bend.ports.values()
    return abs(p0.x - p1.x), abs(p0.y - p1.y)


@gf.cell
def spiral_inner_io(
    N: int = 6,
    x_straight_inner_right: float = 150.0,
    x_straight_inner_left: float = 50.0,
    y_straight_inner_top: float = 50.0,
    y_straight_inner_bottom: float = 10.0,
    grating_spacing: float = 127.0,
    waveguide_spacing: float = 3,
    bend90_function: ComponentFactory = bend_euler,
    bend180_function: ComponentFactory = bend_euler180,
    straight_factory: ComponentFactory = straight,
    taper: Optional[ComponentFactory] = None,
    length: Optional[float] = None,
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> Component:
    """Spiral with ports inside the spiral loop.

    Args:
        N: number of loops
        x_straight_inner_right:
        x_straight_inner_left:
        y_straight_inner_top:
        y_straight_inner_bottom:
        grating_spacing:
        waveguide_spacing: center to center spacing
        bend90_function
        bend180_function
        straight_factory: straight function
        taper: taper function
        length: computes spiral length from simple interpolation
        cross_section:
        **kwargs: cross_section settings

    """
    dx = dy = waveguide_spacing
    cross_section = gf.partial(cross_section, **kwargs)
    x = cross_section()
    width = x.info.get("width")
    taper_length = x.info.get("taper_length", 10.0)

    if length:
        x_straight_inner_left = get_straight_length(
            length=length,
            spiral_function=spiral_inner_io,
            N=N,
            x_straight_inner_right=x_straight_inner_right,
            x_straight_inner_left=x_straight_inner_left,
            y_straight_inner_top=y_straight_inner_top,
            y_straight_inner_bottom=y_straight_inner_bottom,
            grating_spacing=grating_spacing,
            waveguide_spacing=waveguide_spacing,
        )

    _bend180 = gf.call_if_func(bend180_function, cross_section=cross_section)
    _bend90 = gf.call_if_func(bend90_function, cross_section=cross_section)

    rx, ry = get_bend_port_distances(_bend90)
    _, rx180 = get_bend_port_distances(_bend180)  # rx180, second arg since we rotate

    component = gf.Component()
    # gc_port_lbl = 1
    # gc1 = _gc.ref(port_id=gc_port_lbl, position=(0, 0), rotation=-90)
    # gc2 = _gc.ref(port_id=gc_port_lbl, position=(grating_spacing, 0), rotation=-90)
    # component.add([gc1, gc2])

    p1 = gf.Port(
        name="o1",
        midpoint=(0, y_straight_inner_top),
        orientation=270,
        width=width,
        layer=gf.LAYER.WG,
    )
    p2 = gf.Port(
        name="o2",
        midpoint=(grating_spacing, y_straight_inner_top),
        orientation=270,
        width=width,
        layer=gf.LAYER.WG,
    )
    taper = gf.components.taper(
        width1=width,
        width2=_bend180.ports["o1"].width,
        length=taper_length + y_straight_inner_top - 15 - 35,
        cross_section=cross_section,
    )
    taper_ref1 = component.add_ref(taper)
    taper_ref1.connect("o2", p1)

    taper_ref2 = component.add_ref(taper)
    taper_ref2.connect("o2", p2)

    component.absorb(taper_ref1)
    component.absorb(taper_ref2)

    component.add_port(name="o1", port=taper_ref1.ports["o1"])
    component.add_port(name="o2", port=taper_ref2.ports["o1"])

    # Create manhattan path going from west grating to westest port of bend 180
    _pt = np.array(p1.position)
    pts_w = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i + 1) * dy
        x2 = grating_spacing + 2 * rx + x_straight_inner_right + (2 * i + 1) * dx
        y3 = -y_straight_inner_bottom - ry - (2 * i + 3) * dy
        x4 = -x_straight_inner_left - (2 * i + 1) * dx
        if i == N - 1:
            x4 = x4 - rx180 + dx

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_w += [_pt1, _pt2, _pt3, _pt4, _pt5]

    route_west = round_corners(
        pts_w,
        bend_factory=_bend90,
        straight_factory=straight_factory,
        taper=taper,
        cross_section=cross_section,
    )
    component.add(route_west.references)

    # Add loop back
    bend180_ref = _bend180.ref(port_id="o2", position=route_west.ports[1], rotation=90)
    component.add(bend180_ref)
    component.absorb(bend180_ref)

    # Create manhattan path going from east grating to eastest port of bend 180
    _pt = np.array(p2.position)
    pts_e = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i) * dy
        x2 = grating_spacing + 2 * rx + x_straight_inner_right + 2 * i * dx
        y3 = -y_straight_inner_bottom - ry - (2 * i + 2) * dy
        x4 = -x_straight_inner_left - (2 * i) * dx

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_e += [_pt1, _pt2, _pt3, _pt4, _pt5]

    route_east = round_corners(
        pts_e,
        bend_factory=_bend90,
        straight_factory=straight_factory,
        taper=taper,
        cross_section=cross_section,
    )
    component.add(route_east.references)

    length = route_east.length + route_west.length + _bend180.length
    component.length = snap_to_grid(length + 2 * y_straight_inner_top)
    return component


def get_straight_length(
    length: float, spiral_function: ComponentFactory, **kwargs
) -> float:
    """Returns y_spiral to achieve a particular spiral length"""
    x0 = 50
    x1 = 400
    kwargs.update({"x_straight_inner_left": x0})
    s0 = spiral_function(**kwargs)
    kwargs.update({"x_straight_inner_left": x1})
    s1 = spiral_function(**kwargs)
    p = np.polyfit(np.array([x0, x1]), np.array([s0.length, s1.length]), deg=1)
    x_straight_inner_left = (length - p[1]) / p[0]
    # print(x_straight_inner_left)
    return x_straight_inner_left


if __name__ == "__main__":
    # c = spiral_inner_io(radius=20, width=0.2)
    c = spiral_inner_io(radius=40, width=2.0, length=15e3)
    c.show()



from typing import Optional, Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.tech import LAYER


@gf.cell
def via(
    width: float = 0.7,
    height: Optional[float] = None,
    pitch: float = 2.0,
    pitch_x: Optional[float] = None,
    pitch_y: Optional[float] = None,
    enclosure: float = 1.0,
    layer: Tuple[int, int] = LAYER.VIA1,
    layers_cladding: Optional[Tuple[Tuple[int, int], ...]] = None,
    cladding_offset: float = 0,
) -> Component:
    """Rectangular via. Defaults to a square via.

    Args:
        width: in x direction
        height: in y direction, defaults to width
        pitch:
        pitch_x: Optional x pitch
        pitch_y: Optional y pitch
        enclosure: inclusion of via
        layer: via layer
    """
    height = height or width
    c = Component()
    c.info["pitch"] = pitch
    c.info["pitch_x"] = pitch_x or pitch
    c.info["pitch_y"] = pitch_y or pitch
    c.info["enclosure"] = enclosure
    c.info["width"] = width
    c.info["height"] = height

    a = width / 2
    b = height / 2

    c.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)

    layers_cladding = layers_cladding or []
    a = (width + cladding_offset) / 2
    b = (height + cladding_offset) / 2
    for layer in layers_cladding:
        c.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)

    return c


via1 = gf.partial(via, layer=LAYER.VIA1)
via2 = gf.partial(via, layer=LAYER.VIA2, enclosure=2)
via3 = gf.partial(via, layer=LAYER.VIA3)


if __name__ == "__main__":

    c = via()
    # c.pprint
    print(c)
    c.show()



"""Work with SParameters.
"""

from gdsfactory.sp.get_sparameters_path import get_sparameters_path
from gdsfactory.sp.plot import plot
from gdsfactory.sp.read import read_sparameters_component, read_sparameters_lumerical
from gdsfactory.sp.write import write

__all__ = [
    "get_sparameters_path",
    "plot",
    "read_sparameters_component",
    "read_sparameters_lumerical",
    "write",
]



import pytest
from pytest_regressions.data_regression import DataRegressionFixture
from pytest_regressions.num_regression import NumericRegressionFixture

from gdsfactory.add_grating_couplers import add_grating_couplers
from gdsfactory.add_padding import add_padding_container
from gdsfactory.add_termination import add_gratings_and_loopback, add_termination
from gdsfactory.components.cavity import cavity
from gdsfactory.components.extension import extend_ports

# from gdsfactory.components.straight import straight
# from gdsfactory.components.straight_heater import straight_heater
# from gdsfactory.difftest import difftest
from gdsfactory.components.mmi2x2 import mmi2x2
from gdsfactory.components.mzi_phase_shifter import mzi_phase_shifter
from gdsfactory.components.ring_single_dut import ring_single_dut
from gdsfactory.components.spiral_inner_io import spiral_inner_io
from gdsfactory.rotate import rotate
from gdsfactory.routing.add_electrical_pads_shortest import add_electrical_pads_shortest
from gdsfactory.routing.add_electrical_pads_top import add_electrical_pads_top
from gdsfactory.routing.add_fiber_array import add_fiber_array
from gdsfactory.routing.add_fiber_single import add_fiber_single
from gdsfactory.routing.fanout2x2 import fanout2x2

container_factory = dict(
    add_electrical_pads_shortest=add_electrical_pads_shortest,
    add_electrical_pads_top=add_electrical_pads_top,
    add_fiber_array=add_fiber_array,
    add_fiber_single=add_fiber_single,
    add_grating_couplers=add_grating_couplers,
    add_padding_container=add_padding_container,
    add_termination=add_termination,
    extend_ports=extend_ports,
    fanout2x2=fanout2x2,
    rotate=rotate,
    cavity=cavity,
    ring_single_dut=ring_single_dut,
)

container_names = set(container_factory.keys()) - set()
component = mzi_phase_shifter(splitter=mmi2x2)


@pytest.mark.parametrize("container_type", container_names)
def test_settings(container_type: str, data_regression: DataRegressionFixture) -> None:
    """Avoid regressions when exporting settings."""
    c = container_factory[container_type](component=component)
    data_regression.check(c.get_settings())


@pytest.mark.parametrize("container_type", container_names)
def test_ports(container_type: str, num_regression: NumericRegressionFixture) -> None:
    """Avoid regressions in port names and locations."""
    c = container_factory[container_type](component=component)
    if c.ports:
        num_regression.check(c.get_ports_array())


# @pytest.mark.parametrize("container_type", container_names)
# def test_gds(container_type: str) -> None:
#     """Avoid regressions in GDS geometry shapes and layers."""
#     c = container_factory[container_type](component=component)
#     difftest(c)


# Special test cases for exotic components


def test_add_gratings_and_loopback(data_regression: DataRegressionFixture) -> None:
    """This container requires all ports to face the same direction."""
    c = add_gratings_and_loopback(component=spiral_inner_io())
    data_regression.check(c.get_settings())



from typing import Tuple

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.straight import straight


@cell
def array(
    component: gf.types.ComponentOrFactory = straight,
    spacing: Tuple[float, float] = (150.0, 150.0),
    columns: int = 6,
    rows: int = 1,
) -> Component:
    """Returns an array of components.

    Args:
        component: to replicate
        n: number of components
        pitch: float
        axis: x or y
        rotation: in degrees
        h_mirror: horizontal mirror
        v_mirror: vertical mirror
    """
    c = Component()
    component = component() if callable(component) else component
    c.add_array(component, columns=columns, rows=rows, spacing=spacing)

    for col in range(columns):
        for row in range(rows):
            for port in component.ports.values():
                name = f"{port.name}{row+1}{col+1}"
                c.add_port(name, port=port)
                c.ports[name].move((col * spacing[0], row * spacing[1]))
    return c


if __name__ == "__main__":

    c2 = array()
    c2.show(show_ports=True)



import pytest
from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory.sp as sparameters
from gdsfactory import components

factory = {
    i: getattr(components, i)
    for i in dir(components)
    if not i.startswith("_") and callable(getattr(components, i))
}


component_types = [
    "straight",
    "bend_circular",
    "bend_euler",
    "coupler",
    "mmi1x2",
    "mmi2x2",
]
component_types = []


@pytest.mark.parametrize("component_type", component_types)
def test_sp_load(
    component_type: str, data_regression: DataRegressionFixture, check: bool = True
) -> None:
    c = factory[component_type]()
    sp = sparameters.read_sparameters_component(c)

    port_names = sp[0]
    f = list(sp[1])
    s = sp[2]

    lenf = s.shape[0]
    rows = s.shape[1]
    cols = s.shape[2]

    assert rows == cols == len(c.ports)
    assert len(port_names) == len(c.ports)
    if check:
        data_regression.check(dict(port_names=port_names))
    assert lenf == len(f)


if __name__ == "__main__":
    # c = gf.components.straight(layer=(2, 0))
    # print(c.get_sparameters_path())
    test_sp_load("straight", None, False)



from typing import Tuple, Union

from numpy import array

import gdsfactory as gf

Coordinate = Union[Tuple[float, float], array]


@gf.cell_without_validator
def bbox(
    bbox: Tuple[Coordinate, Coordinate] = ((-1.0, -1.0), (3.0, 4.0)),
    layer: Tuple[int, int] = (1, 0),
) -> gf.Component:
    """Returns bounding box rectangle from coordinates, to allow
    creation of a rectangle bounding box directly from another shape.

    Args:
        bbox: Coordinates of the box [(x1, y1), (x2, y2)].
        layer:

    """
    D = gf.Component()
    (a, b), (c, d) = bbox
    points = ((a, b), (c, b), (c, d), (a, d))
    D.add_polygon(points, layer=layer)
    return D


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.L()
    c << bbox(bbox=c.bbox)
    # c = bbox()
    c.show()



import gdsfactory as gf


def test_path_transition_class():
    P = gf.path.straight(length=10, npoints=101)

    X1 = gf.CrossSection()
    X1.add(width=1, offset=0, layer=gf.LAYER.WG, name="core", ports=("o1", "o2"))
    X1.add(width=3, offset=0, layer=gf.LAYER.SLAB90)

    X2 = gf.CrossSection()
    X2.add(width=3, offset=0, layer=gf.LAYER.WG, name="core", ports=("o1", "o2"))

    T = gf.path.transition(X1, X2)
    c3 = gf.path.extrude(P, T)

    sections = {
        section["name"]: section for section in X1.sections if "name" in section
    }

    assert c3.ports["o1"].name == sections["core"]["ports"][0]
    assert c3.ports["o1"].layer == sections["core"]["layer"]
    assert c3.ports["o1"].orientation == 180
    assert c3.ports["o1"].port_type == sections["core"]["port_types"][0]


def test_path_transition_function():
    P = gf.path.straight(length=10, npoints=101)
    X1 = gf.cross_section.cross_section(width=1)
    X2 = gf.cross_section.cross_section(width=3)
    T = gf.path.transition(X1, X2)
    P = gf.path.straight(length=10, npoints=101)
    c3 = gf.path.extrude(P, T)

    sections = {
        section["name"]: section for section in X1.sections if "name" in section
    }

    assert c3.ports["o1"].name == sections["_default"]["ports"][0]
    assert c3.ports["o1"].layer == sections["_default"]["layer"]
    assert c3.ports["o1"].orientation == 180
    assert c3.ports["o1"].port_type == sections["_default"]["port_types"][0]


if __name__ == "__main__":
    test_path_transition_class()
    # test_path_transition_function()

    # X1 = gf.cross_section.cross_section(width=1)
    # X2 = gf.cross_section.cross_section(width=3)
    # T = gf.path.transition(X1, X2)
    # P = gf.path.straight(length=10, npoints=101)
    # c3 = gf.path.extrude(P, T)

    # sections = {
    #     section["name"]: section for section in X1.sections if "name" in section
    # }

    # assert c3.ports["o1"].name == sections["_default"]["ports"][0]
    # assert c3.ports["o1"].layer == sections["_default"]["layer"]
    # assert c3.ports["o1"].orientation == 180
    # assert c3.ports["o1"].port_type == sections["_default"]["port_types"][0]
    # assert c3.ports["o1"].port_type == X1.port_type



"""Straight Doped PIN waveguide."""
from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.extension import extend_ports
from gdsfactory.components.straight import straight
from gdsfactory.components.taper import taper_strip_to_ridge
from gdsfactory.components.via_stack import via_stack_slab
from gdsfactory.cross_section import pin, pn
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def straight_pin(
    length: float = 500.0,
    cross_section: CrossSectionFactory = pin,
    via_stack: ComponentFactory = via_stack_slab,
    via_stack_width: float = 10.0,
    via_stack_spacing: float = 2,
    port_orientation_top: int = 0,
    port_orientation_bot: int = 180,
    taper: Optional[ComponentFactory] = taper_strip_to_ridge,
    **kwargs,
) -> Component:
    """Returns PIN with contacts

    https://doi.org/10.1364/OE.26.029983

    500um length from
    https://ieeexplore.ieee.org/document/8268112

    Args:
        length: of the waveguide
        cross_section: for the waveguide
        via_stack: for the contacts
        via_stack_size:
        via_stack_spacing: spacing between contacts
        port_orientation_top: for top contact
        port_orientation_bot: for bottom contact
        taper: optional taper
        kwargs: cross_section settings

    """
    c = Component()
    if taper:
        taper = taper() if callable(taper) else taper
        length -= 2 * taper.get_ports_east_west_distance()

    wg = c << gf.c.straight(
        cross_section=cross_section,
        length=length,
        **kwargs,
    )

    if taper:
        t1 = c << taper
        t2 = c << taper
        t1.connect("o2", wg.ports["o1"])
        t2.connect("o2", wg.ports["o2"])
        c.add_port("o1", port=t1.ports["o1"])
        c.add_port("o2", port=t2.ports["o1"])

    else:
        c.add_ports(wg.get_ports_list())

    via_stack_length = length
    contact_top = c << via_stack(
        size=(via_stack_length, via_stack_width),
    )
    contact_bot = c << via_stack(
        size=(via_stack_length, via_stack_width),
    )

    contact_bot.xmin = wg.xmin
    contact_top.xmin = wg.xmin

    contact_top.ymin = +via_stack_spacing / 2
    contact_bot.ymax = -via_stack_spacing / 2

    c.add_port(
        "e1", port=contact_top.get_ports_list(orientation=port_orientation_top)[0]
    )
    c.add_port(
        "e2", port=contact_bot.get_ports_list(orientation=port_orientation_bot)[0]
    )
    return c


straight_pin_passive = gf.partial(straight, cross_section=pin)

straight_pin_passive_tapered = gf.partial(
    extend_ports,
    component=straight_pin_passive,
    extension_factory=taper_strip_to_ridge,
    port1="o2",
    port2="o1",
)

straight_pn_passive = gf.partial(straight, cross_section=pn)
straight_pn_passive_tapered = gf.partial(
    extend_ports,
    component=straight_pn_passive,
    extension_factory=taper_strip_to_ridge,
    port1="o2",
    port2="o1",
)


straight_pn = gf.partial(straight_pin, cross_section=pn)

if __name__ == "__main__":
    c = straight_pin()
    # c = straight_pn_passive_tapered()
    # print(c.ports.keys())
    c.show()



"""pack a list of components into as few components as possible.
adapted from phidl.geometry.
"""

from typing import Any, Dict, List, Tuple, Union

import numpy as np
import rectpack

from gdsfactory.component import Component
from gdsfactory.types import Coordinate, Number


def _pack_single_bin(
    rect_dict: Dict[int, Tuple[Number, Number]],
    aspect_ratio: Tuple[Number, Number],
    max_size: Tuple[float, float],
    sort_by_area: bool,
    density: float,
    precision: float,
) -> Tuple[Dict[int, Tuple[Number, Number, Number, Number]], Dict[Any, Any]]:
    """Packs a dict of rectangles {id:(w,h)} and tries to
    pack it into a bin as small as possible with aspect ratio `aspect_ratio`
    Will iteratively grow the bin size until everything fits or the bin size
    reaches `max_size`.

    Args:
        rect_dict: dict of rectangles {id: (w, h)} to pack
        aspect_ratio:
        max_size: tuple of max X, Y size
        sort_by_area: sorts components by area
        density: of packing. Values closer to 1 require more computation to pack tighter
        precision: Desired precision for rounding vertex coordinates.

    Returns:
        packed rectangles dict {id:(x,y,w,h)}
        dict of remaining unpacked rectangles
    """

    # Compute total area and use it for an initial estimate of the bin size
    total_area = 0
    for r in rect_dict.values():
        total_area += r[0] * r[1]
    aspect_ratio = np.asarray(aspect_ratio) / np.linalg.norm(aspect_ratio)  # Normalize

    # Setup variables
    box_size = np.asarray(aspect_ratio * np.sqrt(total_area), dtype=np.float64)
    box_size = np.clip(box_size, None, max_size)
    if sort_by_area:
        rp_sort = rectpack.SORT_AREA
    else:
        rp_sort = rectpack.SORT_NONE

    # Repeatedly run the rectangle-packing algorithm with increasingly larger
    # areas until everything fits or we've reached the maximum size
    while True:
        # Create the pack object
        rect_packer = rectpack.newPacker(
            mode=rectpack.PackingMode.Offline,
            pack_algo=rectpack.MaxRectsBlsf,
            sort_algo=rp_sort,
            bin_algo=rectpack.PackingBin.BBF,
            rotation=False,
        )

        # Add each rectangle to the pack, create a single bin, and pack
        for rid, r in rect_dict.items():
            rect_packer.add_rect(width=r[0], height=r[1], rid=rid)
        rect_packer.add_bin(width=box_size[0], height=box_size[1])
        rect_packer.pack()

        # Adjust the box size for next time
        box_size *= density  # Increase area to try to fit
        box_size = np.clip(box_size, None, max_size)

        # Quit the loop if we've packed all the rectangles or reached the max size
        if len(rect_packer.rect_list()) == len(rect_dict):
            break
        if all(box_size >= max_size):
            break

    # Separate packed from unpacked rectangles, make dicts of form {id:(x,y,w,h)}
    packed_rect_dict = {r[-1]: r[:-1] for r in rect_packer[0].rect_list()}
    unpacked_rect_dict = {}
    for k, v in rect_dict.items():
        if k not in packed_rect_dict:
            unpacked_rect_dict[k] = v

    return (packed_rect_dict, unpacked_rect_dict)


def pack(
    component_list: List[Component],
    spacing: float = 10.0,
    aspect_ratio: Tuple[Number, Number] = (1, 1),
    max_size: Union[Coordinate, Tuple[None, None]] = (None, None),
    sort_by_area: bool = True,
    density: float = 1.1,
    precision: float = 1e-2,
) -> List[Component]:
    """Pack a list of components into as few Components as possible.
    Adapted from phidl.geometry

    Args:
        component_list: Must be a list or tuple of Components
        spacing: Minimum distance between adjacent shapes
        aspect_ratio: (width, height) ratio of the rectangular bin
        max_size: Limits the size into which the shapes will be packed
        sort_by_area: Pre-sorts the shapes by area
        density: Values closer to 1 pack tighter but require more computation
        precision: Desired precision for rounding vertex coordinates.
    """

    if density < 1.01:
        raise ValueError(
            "pack() was given a `density` argument that is"
            + " too small.  The density argument must be >= 1.01"
        )

    # Santize max_size variable
    max_size = [np.inf if v is None else v for v in max_size]
    max_size = np.asarray(max_size, dtype=np.float64)  # In case it's integers
    max_size = max_size / precision

    # Convert Components to rectangles
    rect_dict = {}
    for n, D in enumerate(component_list):
        w, h = (D.size + spacing) / precision
        w, h = int(w), int(h)
        if (w > max_size[0]) or (h > max_size[1]):
            raise ValueError(
                "pack() failed because one of the objects (D)"
                + "in `component_list` is has an x or y dimension larger than `max_size` and "
                + "so cannot be packed"
            )
        rect_dict[n] = (w, h)

    packed_list = []
    while len(rect_dict) > 0:
        (packed_rect_dict, rect_dict) = _pack_single_bin(
            rect_dict,
            aspect_ratio=aspect_ratio,
            max_size=max_size,
            sort_by_area=sort_by_area,
            density=density,
            precision=precision,
        )
        packed_list.append(packed_rect_dict)

    components_packed_list = []
    for rect_dict in packed_list:
        packed = Component()
        packed.settings["components"] = {}
        for n, rect in rect_dict.items():
            x, y, w, h = rect
            xcenter = x + w / 2 + spacing / 2
            ycenter = y + h / 2 + spacing / 2
            component = component_list[n]
            d = packed.add_ref(component)
            if hasattr(component, "settings"):
                packed.settings["components"][component.name] = component.get_settings()
            d.center = (xcenter * precision, ycenter * precision)
        components_packed_list.append(packed)

    return components_packed_list


def test_pack() -> Component:
    import phidl.geometry as pg

    component_list = [pg.ellipse(radii=np.random.rand(2) * n + 2) for n in range(2)]
    component_list += [pg.rectangle(size=np.random.rand(2) * n + 2) for n in range(2)]

    components_packed_list = pack(
        component_list,  # Must be a list or tuple of Components
        spacing=1.25,  # Minimum distance between adjacent shapes
        aspect_ratio=(2, 1),  # (width, height) ratio of the rectangular bin
        max_size=(None, None),  # Limits the size into which the shapes will be packed
        density=1.05,  # Values closer to 1 pack tighter but require more computation
        sort_by_area=True,  # Pre-sorts the shapes by area
    )
    c = components_packed_list[0]  # Only one bin was created, so we plot that
    # print(len(c.get_dependencies()))
    assert len(c.get_dependencies()) == 4
    return c


def test_pack_with_settings() -> Component:
    import gdsfactory as gf

    component_list = [gf.components.rectangle(size=(i, i)) for i in range(1, 10)]
    component_list += [gf.components.rectangle(size=(i, i)) for i in range(1, 10)]

    components_packed_list = pack(
        component_list,  # Must be a list or tuple of Components
        spacing=1.25,  # Minimum distance between adjacent shapes
        aspect_ratio=(2, 1),  # (width, height) ratio of the rectangular bin
        max_size=(None, None),  # Limits the size into which the shapes will be packed
        density=1.05,  # Values closer to 1 pack tighter but require more computation
        sort_by_area=True,  # Pre-sorts the shapes by area
        precision=1e-3,
    )
    c = components_packed_list[0]  # Only one bin was created, so we plot that
    # print(len(c.get_dependencies()))
    return c


if __name__ == "__main__":

    c = test_pack_with_settings()
    # c = test_pack()
    c.show()
    c.pprint
    c.write_gds_with_metadata("mask.gds")

    # import phidl.geometry as pg
    # spacing = 1
    # ellipses = pack(
    #     [pg.ellipse(radii=np.random.rand(2) * n + 2) for n in range(50)],
    #     spacing=spacing,
    # )[0]
    # ellipses.name = "ellipses"
    # rectangles = pack(
    #     [pg.rectangle(size=np.random.rand(2) * n + 2) for n in range(50)],
    #     spacing=spacing,
    # )[0]
    # rectangles.name = "rectangles"
    # p = pack([ellipses, rectangles])
    # p.show()



from typing import Optional

import numpy as np
import picwriter.components as pc

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.waveguide_template import strip
from gdsfactory.port import deco_rename_ports
from gdsfactory.types import ComponentFactory, Coordinate, Coordinates, Layer


@deco_rename_ports
@cell
def grating_coupler_elliptical2(
    wgt: ComponentFactory = strip,
    wg_width: float = 0.5,
    cladding_offset: float = 2.0,
    theta: float = np.pi / 4.0,
    length: float = 30.0,
    taper_length: float = 10.0,
    period: float = 1.0,
    dutycycle: float = 0.7,
    port: Coordinate = (0.0, 0.0),
    layer_ridge: Optional[Layer] = None,
    layer_core: Layer = gf.LAYER.WG,
    layer_cladding: Layer = gf.LAYER.WGCLAD,
    teeth_list: Optional[Coordinates] = None,
    direction: str = "EAST",
    polarization: str = "te",
    wavelength_nm: float = 1550.0,
    fiber_marker_width: float = 11.0,
    fiber_marker_layer: Layer = gf.LAYER.TE,
    **kwargs,
) -> Component:
    r"""Returns Grating coupler from Picwriter

    Args:
        wgt: waveguide_template object or function
        theta: Angle of the straight in rad.
        length: total grating coupler region.
        taper_length: Length of the taper before the grating coupler.
        period: Grating period.
        dutycycle: (period-gap)/period.
        port: Cartesian coordinate of the input port
        layer_ridge: for partial etched gratings
        layer_core: Tuple specifying the layer/datatype of the ridge region.
        layer_cladding: for the straight.
        teeth_list: (gap, width) tuples to be used as the gap and teeth widths
            for irregularly spaced gratings.
            For example, [(0.6, 0.2), (0.7, 0.3), ...] would be a gap of 0.6,
            then a tooth of width 0.2, then gap of 0.7 and tooth of 0.3, and so on.
            Overrides *period*, *dutycycle*, and *length*.  Defaults to None.
        direction: Direction that the component will point *towards*,
            can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`,
            OR an angle (float, in radians)
        polarization: te or tm
        wavelength_nm: wavelength in nm

    .. plot::
      :include-source:

      import gdsfactory as gf

      c = gf.components.grating_coupler_elliptical2()
      c.plot()

    .. code::

                      fiber

                   /  /  /  /
                  /  /  /  /
                _|-|_|-|_|-|___
        WG->W0  ______________|

    """
    ridge = True if layer_ridge else False

    c = pc.GratingCoupler(
        gf.call_if_func(
            wgt,
            cladding_offset=cladding_offset,
            wg_width=wg_width,
            layer=layer_core,
            layer_cladding=layer_cladding,
            **kwargs,
        ),
        theta=theta,
        length=length,
        taper_length=taper_length,
        period=period,
        dutycycle=1 - dutycycle,
        ridge=ridge,
        ridge_layers=layer_ridge,
        teeth_list=teeth_list,
        port=port,
        direction=direction,
    )

    c = gf.component_from.picwriter(c)
    c.polarization = polarization
    c.wavelength = wavelength_nm

    x = c.center[0] + taper_length / 2
    circle = gf.components.circle(
        radius=fiber_marker_width / 2, layer=fiber_marker_layer
    )
    circle_ref = c.add_ref(circle)
    circle_ref.movex(x)

    c.add_port(
        name=f"vertical_{polarization.lower()}",
        midpoint=[x, 0],
        width=fiber_marker_width,
        orientation=0,
        layer=fiber_marker_layer,
    )

    return c


if __name__ == "__main__":

    c = grating_coupler_elliptical2(length=31, taper_length=30)
    print(len(c.name))
    print(c.ports)
    c.show()



"""Read component GDS, JSON metadata and CSV ports."""

import gdsfactory as gf
from gdsfactory.component import Component


def remove_labels(component: Component, layer=gf.LAYER.LABEL_SETTINGS) -> None:
    """Returns same component without labels."""
    for c in list(component.get_dependencies(recursive=True)) + [component]:
        old_label = [
            label for label in c.labels if label.layer == gf.LAYER.LABEL_SETTINGS
        ]
        if len(old_label) > 0:
            for label in old_label:
                c.labels.remove(label)


if __name__ == "__main__":
    c = gf.components.waveguide()
    gf.show()



from numpy import float64

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bend_circular import bend_circular, bend_circular180
from gdsfactory.components.bend_euler import bend_euler, bend_euler180
from gdsfactory.components.component_sequence import component_sequence
from gdsfactory.components.straight import straight
from gdsfactory.types import ComponentFactory, ComponentOrFactory


def _get_bend_size(bend90: Component) -> float64:
    p1, p2 = list(bend90.ports.values())[:2]
    bsx = abs(p2.x - p1.x)
    bsy = abs(p2.y - p1.y)
    return max(bsx, bsy)


@cell
def cutback_bend(
    bend90: ComponentOrFactory = bend_euler,
    straight_length: float = 5.0,
    n_steps: int = 6,
    n_stairs: int = 5,
):
    """Deprecated! use cutback_bend90 instead!
    this is a stair

    .. code::
            _
          _|
        _|

        _ this is a step

    """

    bend90 = gf.call_if_func(bend90)
    wg = straight(length=straight_length, width=bend90.ports["o1"].width)

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component = {
        "A": (bend90, "o1", "o2"),
        "B": (bend90, "o2", "o1"),
        "S": (wg, "o1", "o2"),
    }

    # Generate the sequence of staircases
    s = ""
    for i in range(n_stairs):
        s += "ASBS" * n_steps
        s += "ASAS" if i % 2 == 0 else "BSBS"
    s = s[:-4]

    c = component_sequence(
        sequence=s, symbol_to_component=symbol_to_component, start_orientation=90
    )
    c.update_settings(n_bends=n_steps * n_stairs * 2 + n_stairs * 2 - 2)
    return c


@cell
def cutback_bend90(
    bend90: ComponentOrFactory = bend_euler,
    straight_length: float = 5.0,
    n_steps: int = 6,
    cols: int = 6,
    spacing: int = 5,
    wg_loop_length: None = None,
    straight_factory: ComponentFactory = straight,
) -> Component:
    """

    .. code::

           _
        |_| |

    """
    bend90 = gf.call_if_func(bend90)

    wg = straight_factory(length=straight_length, width=bend90.ports["o1"].width)
    if wg_loop_length is None:
        wg_loop_length = 2 * _get_bend_size(bend90) + spacing + straight_length

    wg_loop = straight_factory(
        length=wg_loop_length,
        width=bend90.ports["o1"].width,
    )
    # Define a map between symbols and (component, input port, output port)
    symbol_to_component = {
        "A": (bend90, "o1", "o2"),
        "B": (bend90, "o2", "o1"),
        "-": (wg, "o1", "o2"),
        "|": (wg_loop, "o1", "o2"),
    }

    # Generate the sequence of staircases
    s = ""
    for i in range(cols):
        if i % 2 == 0:  # even row
            s += "A-A-B-B-" * n_steps + "|"
        else:
            s += "B-B-A-A-" * n_steps + "|"
    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(
        sequence=s, symbol_to_component=symbol_to_component, start_orientation=0
    )
    c.update_settings(n_bends=n_steps * cols * 4)
    return c


@cell
def staircase(
    bend90: ComponentOrFactory = bend_euler,
    length_v: float = 5.0,
    length_h: float = 5.0,
    n_steps: int = 4,
    straight_factory: ComponentFactory = straight,
) -> Component:
    bend90 = gf.call_if_func(bend90)

    wgh = straight_factory(length=length_h, width=bend90.ports["o1"].width)
    wgv = straight_factory(length=length_v, width=bend90.ports["o1"].width)

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component = {
        "A": (bend90, "o1", "o2"),
        "B": (bend90, "o2", "o1"),
        "-": (wgh, "o1", "o2"),
        "|": (wgv, "o1", "o2"),
    }

    # Generate the sequence of staircases
    s = "-A|B" * n_steps + "-"

    c = component_sequence(
        sequence=s, symbol_to_component=symbol_to_component, start_orientation=0
    )
    c.update_settings(n_bends=2 * n_steps)
    return c


@cell
def cutback_bend180(
    bend180: ComponentOrFactory = bend_euler180,
    straight_length: float = 5.0,
    n_steps: int = 6,
    cols: int = 6,
    spacing: int = 3,
    straight_factory: ComponentFactory = straight,
) -> Component:
    """

    .. code::

          _
        _| |_| this is a stair

        _ this is a step

    """
    bend180 = gf.call_if_func(bend180)

    wg = straight_factory(length=straight_length, width=bend180.ports["o1"].width)
    wg_vertical = straight_factory(
        length=2 * bend180.size_info.width + straight_length + spacing,
        width=bend180.ports["o1"].width,
    )

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component = {
        "D": (bend180, "o1", "o2"),
        "C": (bend180, "o2", "o1"),
        "-": (wg, "o1", "o2"),
        "|": (wg_vertical, "o1", "o2"),
    }

    # Generate the sequence of staircases
    s = ""
    for i in range(cols):
        if i % 2 == 0:  # even row
            s += "D-C-" * n_steps + "|"
        else:
            s += "C-D-" * n_steps + "|"
    s = s[:-1]

    # Create the component from the sequence
    c = component_sequence(
        sequence=s, symbol_to_component=symbol_to_component, start_orientation=0
    )
    c.update_settings(n_bends=n_steps * cols * 2 + cols * 2 - 2)
    return c


cutback_bend180circular = gf.partial(cutback_bend180, bend180=bend_circular180)
cutback_bend90circular = gf.partial(cutback_bend90, bend90=bend_circular)

if __name__ == "__main__":
    # c = cutback_bend_circular(n_steps=7, n_stairs=4, radius=5) #62
    # c = cutback_bend_circular(n_steps=14, n_stairs=4) #118
    # c = cutback_bend90()
    c = cutback_bend180()
    c.show()



"""Basic checks."""

from gdsfactory.drc.check_duplicate_cells import check_duplicate_cells
from gdsfactory.drc.check_exclusion import check_exclusion
from gdsfactory.drc.check_inclusion import check_inclusion
from gdsfactory.drc.check_space import check_space
from gdsfactory.drc.check_width import check_width
from gdsfactory.drc.density import compute_area

__all__ = [
    "check_space",
    "check_width",
    "check_exclusion",
    "check_inclusion",
    "check_duplicate_cells",
    "compute_area",
]



"""Read component GDS, JSON metadata and CSV ports."""
import csv
import json
from pathlib import Path
from typing import Union

from gdsfactory.component import Component


def read_ports(component: Component, gdspath: Union[str, Path]) -> None:
    """Adds Component with ports (CSV) and metadata (JSON) info (if any)."""

    if not gdspath.exists():
        raise FileNotFoundError(f"No such file '{gdspath}'")

    ports_filepath = gdspath.with_suffix(".ports")
    metadata_filepath = gdspath.with_suffix(".json")

    if ports_filepath.exists():
        with open(str(ports_filepath), newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=",", quotechar="|")
            for r in reader:
                layer_type = int(r[5].strip().strip("("))
                data_type = int(r[6].strip().strip(")"))
                component.add_port(
                    name=r[0],
                    midpoint=[float(r[1]), float(r[2])],
                    orientation=int(r[3]),
                    width=float(r[4]),
                    layer=(layer_type, data_type),
                )

    if metadata_filepath.exists():
        with open(metadata_filepath) as f:
            data = json.load(f)
        cell_settings = data["cells"][component.name]
        component.settings.update(cell_settings)


if __name__ == "__main__":
    from gdsfactory.tests.test_load_component import test_load_component_gds

    test_load_component_gds()



from typing import Optional

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.array import array
from gdsfactory.components.pad import pad
from gdsfactory.components.straight import straight
from gdsfactory.components.via_stack import via_stack
from gdsfactory.cross_section import metal2
from gdsfactory.types import ComponentFactory, ComponentOrFactory, CrossSectionFactory


@cell
def array_with_via(
    component: ComponentOrFactory = pad,
    n: int = 3,
    pitch: float = 150.0,
    waveguide_pitch: float = 10.0,
    end_straight: float = 60.0,
    component_port_name: str = "e4",
    cross_section: CrossSectionFactory = metal2,
    via_stack: ComponentFactory = via_stack,
    via_stack_y_offset: float = -44.0,
    facing_west: bool = True,
    **kwargs,
) -> Component:
    """Returns an array of components in X axis
    with fanout waveguides facing west

    Args:
        component: to replicate
        n: number of components
        pitch: float
        waveguide_pitch: for fanout
        end_straight: lenght of the straight at the end
        waveguide: waveguide definition
        component_port_name:
        via_stack_port_name:
        **kwargs
    """
    port_orientation = 180 if facing_west else 0

    c = Component()
    component = component() if callable(component) else component
    via_stack = via_stack()

    for col in range(n):
        ref = component.ref()
        ref.x = col * pitch
        c.add(ref)

        if port_orientation == 180:
            xlength = col * pitch + end_straight
        else:
            xlength = n * pitch - (col * pitch) + end_straight

        via_stack_ref = c << via_stack
        via_stack_ref.x = col * pitch
        via_stack_ref.y = col * waveguide_pitch + via_stack_y_offset

        straightx_ref = c << straight(
            length=xlength, cross_section=cross_section, **kwargs
        )
        straightx_ref.connect(
            "e2", via_stack_ref.get_ports_list(orientation=port_orientation)[0]
        )
        c.add_port(f"e{col}", port=straightx_ref.ports["e1"])
    return c


@cell
def array_with_via_2d(
    pitch: float = 150.0,
    pitch_x: Optional[float] = None,
    pitch_y: Optional[float] = None,
    cols: int = 3,
    rows: int = 2,
    **kwargs,
) -> Component:
    """Returns 2D array with fanout waveguides facing west.

    Args:
        pitch: 2D pitch
        pitch_x: defaults to pitch
        pitch_y: defaults to pitch
        cols:
        rows:
        kwargs:
            component: to replicate
            n: number of components
            pitch: float
            waveguide_pitch: for fanout
            end_straight: lenght of the straight at the end
            component_port_name:
            via_stack_port_name:
            **kwargs
    """
    pitch_y = pitch_y or pitch
    pitch_x = pitch_x or pitch
    row = array_with_via(n=cols, pitch=pitch_x, **kwargs)
    return array(component=row, rows=rows, spacing=(0, pitch_y))


if __name__ == "__main__":
    c2 = array_with_via(n=3, width=10, waveguide_pitch=20)
    c2 = array_with_via_2d(cols=8, rows=8, waveguide_pitch=12, facing_west=True)
    c2.show()



from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.tech import LAYER


@gf.cell
def cross(
    length: float = 10.0,
    width: float = 3.0,
    layer: Tuple[int, int] = LAYER.WG,
) -> Component:
    """Generates a right-angle cross from two rectangles of specified length and width.

    Args:
        length: float Length of the cross from one end to the other
        width: float Width of the arms of the cross
        layer: int, array-like[2], or set Specific layer(s) to put polygon geometry on

    """

    c = gf.Component()
    R = gf.components.rectangle(size=(width, length), layer=layer)
    r1 = c.add_ref(R).rotate(90)
    r2 = c.add_ref(R)
    r1.center = (0, 0)
    r2.center = (0, 0)
    c.add_port(
        1,
        width=width,
        layer=layer,
        orientation=0,
        midpoint=(+length / 2, 0),
    )
    c.add_port(
        2,
        width=width,
        layer=layer,
        orientation=180,
        midpoint=(-length / 2, 0),
    )
    c.add_port(
        3,
        width=width,
        layer=layer,
        orientation=90,
        midpoint=(0, length / 2),
    )
    c.add_port(
        4,
        width=width,
        layer=layer,
        orientation=270,
        midpoint=(0, -length / 2),
    )
    c.absorb(r1)
    c.absorb(r2)
    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    c = cross()
    c.show()
    c.pprint_ports
    cc = gf.routing.add_fiber_array(component=c)
    cc.show()



from pathlib import Path
from typing import Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.sp.write import write


def plot(
    component_or_df: Union[Component, DataFrame],
    logscale: bool = True,
    keys: Optional[Tuple[str, ...]] = None,
    dirpath: Path = gf.CONFIG["sp"],
    **sim_settings,
):
    """Plots Sparameters.

    Args:
        component_or_df: Component or Sparameters in pandas DataFrame
        logscale: plots 20*log10(results)
        keys: list of keys to plot
        dirpath: where to store the simulations
        **sim_settings: simulation kwargs

    """

    r = component_or_df
    if isinstance(r, Component):
        r = write(component=r, dirpath=dirpath, **sim_settings)
    w = r["wavelength_nm"]

    if keys:
        keys = [key for key in keys if key in r.keys()]
    else:
        keys = [key for key in r.keys() if key.startswith("S") and key.endswith("m")]

    for key in keys:
        y = 20 * np.log10(r[key]) if logscale else r[key]
        plt.plot(w, y, label=key[:-1])
    plt.legend()
    plt.xlabel("wavelength (nm)")
    plt.ylabel("Transmission (dB)") if logscale else plt.ylabel("Transmission")


if __name__ == "__main__":

    remove_layers = []
    layer_to_thickness_nm = {(1, 0): 220}

    # r = write(component=gf.components.straight(), layer_to_thickness_nm=layer_to_thickness_nm)
    # r = write(component=gf.components.mmi2x2(), layer_to_thickness_nm=layer_to_thickness_nm)
    # r = write(component=gf.components.mmi1x2(), layer_to_thickness_nm=layer_to_thickness_nm)
    # r = write(component=gf.components.coupler(), layer_to_thickness_nm=layer_to_thickness_nm)
    # r = write(component=gf.components.bend_circular(), layer_to_thickness_nm=layer_to_thickness_nm)
    # plot(r, logscale=True)
    # plot(gf.components.coupler())
    plot(gf.components.mmi1x2(), logscale=False)
    plt.show()



"""Add grating_couplers to a component."""
from typing import Callable, List, Tuple

from phidl.device_layout import Label

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.grating_coupler_elliptical_trenches import grating_coupler_te
from gdsfactory.port import select_ports_optical
from gdsfactory.routing.get_input_labels import get_input_labels
from gdsfactory.types import ComponentFactory


@cell
def add_grating_couplers(
    component: Component,
    grating_coupler: ComponentFactory = grating_coupler_te,
    layer_label: Tuple[int, int] = gf.LAYER.LABEL,
    gc_port_name: str = "o1",
    get_input_labels_function: Callable[..., List[Label]] = get_input_labels,
    select_ports: Callable = select_ports_optical,
) -> Component:
    """Returns new component with grating couplers and labels.

    Args:
        component: to add grating_couplers
        grating_coupler: grating_coupler function
        layer_label: for label
        gc_port_name: where to add label
        get_input_labels_function: function to get label
        select_ports: for selecting optical_ports

    """

    cnew = Component()
    cnew.component = component
    cnew.add_ref(component)
    grating_coupler = gf.call_if_func(grating_coupler)

    io_gratings = []
    optical_ports = select_ports(component.ports)
    optical_ports = list(optical_ports.values())
    for port in optical_ports:
        gc_ref = grating_coupler.ref()
        gc_port = gc_ref.ports[gc_port_name]
        gc_ref.connect(gc_port, port)
        io_gratings.append(gc_ref)
        cnew.add(gc_ref)

    labels = get_input_labels_function(
        io_gratings,
        list(component.ports.values()),
        component_name=component.name,
        layer_label=layer_label,
        gc_port_name=gc_port_name,
    )
    cnew.add(labels)
    return cnew


if __name__ == "__main__":
    # from gdsfactory.add_labels import get_optical_text
    # c = gf.components.grating_coupler_elliptical_te()
    # print(c.wavelength)

    # print(c.get_property('wavelength'))

    c = gf.components.straight(width=2)
    cc = add_grating_couplers(component=c)
    cc.show()



from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper import taper_strip_to_ridge
from gdsfactory.components.via_stack import via_stack_metal
from gdsfactory.cross_section import rib_heater_doped, rib_heater_doped_contact
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def straight_heater_doped(
    length: float = 320.0,
    nsections: int = 3,
    cross_section_heater: CrossSectionFactory = rib_heater_doped,
    cross_section_contact: CrossSectionFactory = rib_heater_doped_contact,
    via_stack: ComponentFactory = via_stack_metal,
    via_stack_size: Tuple[float, float] = (10.0, 10.0),
    port_orientation_top: int = 0,
    port_orientation_bot: int = 180,
    taper: ComponentFactory = taper_strip_to_ridge,
    inclusion_contact: float = 0.2,
    **kwargs,
) -> Component:
    r"""Returns a doped thermal phase shifter.
    dimensions from https://doi.org/10.1364/OE.27.010456

    Args:
        length: of the waveguide
        nsections: between contacts
        cross_section_heater: for heated sections
        via_stack: for connecting the metal
        via_stack_size:
        port_orientation_top: for top contact
        port_orientation_bot: for bottom contact
        taper: optional taper
        kwargs: cross_section settings

    .. code::

                              length
         <-------------------------------------------->
                       length_section
             <-------------------------------->
           length_contact
             <------->
             ______________________________________
           /|        |____________________|        |
          / |viastack|                    |        |
          \ | size   |____________________|        |
           \|________|____________________|________|

        taper         cross_section_heater cross_section_contact
    """
    c = Component()

    taper = taper()
    length -= 2 * taper.get_ports_east_west_distance()

    length_contact = via_stack_size[1] - 2 * inclusion_contact
    length_section = (length - length_contact) / nsections
    length_heater = length_section - length_contact

    if length_heater < 0:
        raise ValueError("You need to increase the length")

    wg0 = c << gf.c.straight(
        cross_section=cross_section_heater,
        length=inclusion_contact,
        **kwargs,
    )

    wg1 = c << gf.c.straight(
        cross_section=cross_section_contact,
        length=length_contact - inclusion_contact,
        **kwargs,
    )
    wg1.connect("o1", wg0.ports["o2"])
    x0 = wg0.get_ports_list()[0].x + via_stack_size[0] / 2

    for _ in range(nsections):
        wg2 = c << gf.c.straight(
            cross_section=cross_section_heater,
            length=length_heater,
            **kwargs,
        )
        wg2.connect("o1", wg1.ports["o2"])
        wg3 = c << gf.c.straight(
            cross_section=cross_section_contact,
            length=length_contact,
            **kwargs,
        )
        wg3.connect("o1", wg2.ports["o2"])
        wg1 = wg3

    contact_section = via_stack(size=via_stack_size)
    contacts = []
    for i in range(0, nsections + 1):
        xi = x0 + length_section * i - inclusion_contact
        contact_i_center = c.add_ref(contact_section)
        contact_i_center.x = xi

        contact_i = c << contact_section
        contact_i.x = xi
        contact_i.y = +via_stack_size[1] if i % 2 == 0 else -via_stack_size[1]
        contacts.append(contact_i)

    t1 = c << taper
    t2 = c << taper
    t1.connect("o2", wg0.ports["o1"])
    t2.connect("o2", wg1.ports["o2"])
    c.add_port("o1", port=t1.ports["o1"])
    c.add_port("o2", port=t2.ports["o1"])

    via_stack_length = length + via_stack_size[0]
    contact_top = c << via_stack(
        size=(via_stack_length, via_stack_size[0]),
    )
    contact_bot = c << via_stack(
        size=(via_stack_length, via_stack_size[0]),
    )

    contact_bot.xmin = contacts[0].xmin
    contact_top.xmin = contacts[0].xmin

    contact_top.ymin = contacts[0].ymax
    contact_bot.ymax = contacts[1].ymin

    c.add_port(
        "e1", port=contact_top.get_ports_list(orientation=port_orientation_top)[0]
    )
    c.add_port(
        "e2", port=contact_bot.get_ports_list(orientation=port_orientation_bot)[0]
    )
    return c


def test_straight_heater_doped_ports() -> Component:
    c = straight_heater_doped(length=100.0)
    assert (
        c.get_ports_east_west_distance(port_type="optical") == 100.0
    ), c.get_ports_east_west_distance(port_type="optical")
    return c


if __name__ == "__main__":
    c = test_straight_heater_doped_ports()
    # c = straight_heater_doped(length=80)
    c.show()



import pathlib
from typing import Iterable

from gdsfactory.component import Component
from gdsfactory.import_gds import import_gds
from gdsfactory.types import ComponentOrPath, PathType


def gdspaths(cells: Iterable[ComponentOrPath]) -> Component:
    """Combine all GDS files or gf.components into a gf.component.

    Args:
        cells: List of gdspaths or Components
    """
    c = Component()

    for cell in cells:
        if not isinstance(cell, Component):
            cell = import_gds(cell)
        c << cell

    return c


def gdsdir(dirpath: PathType) -> Component:
    """Merges GDS cells from a directory into a single Component"""
    dirpath = pathlib.Path(dirpath)
    return gdspaths(dirpath.glob("*.gds"))


if __name__ == "__main__":
    from gdsfactory.config import diff_path

    # c = gdspaths([gf.components.straight(), gf.components.bend_circular()])
    # leave these two lines to end up tests showing the diff
    c = gdspaths(diff_path.glob("*.gds"))
    c.show()



"""This example shows how to add_grating couplers for single fiber in single fiber out (no fiber array).
"""

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.samples.big_device import big_device


def test_fiber_single() -> Component:
    w = h = 18 * 50
    c = big_device(port_pitch=50.0, h=h, w=w)
    return gf.routing.add_fiber_single(component=c)


def demo_needs_fix():
    """FIXME"""
    c = big_device()
    return gf.routing.add_fiber_single(component=c)


if __name__ == "__main__":
    c = test_fiber_single()
    # c = demo_needs_fix()
    c.show()



from typing import List, Optional, Union

import phidl.geometry as pg

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.types import ComponentOrReference, Layer


def boolean(
    A: Union[ComponentOrReference, List[ComponentOrReference]],
    B: Union[ComponentOrReference, List[ComponentOrReference]],
    operation: str,
    precision: float = 1e-4,
    num_divisions: Optional[int] = None,
    max_points: int = 4000,
    layer: Layer = (1, 0),
) -> Component:
    """Performs boolean operations between 2 Device/DeviceReference objects,
    or lists of Devices/DeviceReferences.

    ``operation`` should be one of {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}.
    Note that 'A+B' is equivalent to 'or', 'A-B' is equivalent to 'not', and
    'B-A' is equivalent to 'not' with the operands switched

    Args:
        A : Device(/Reference) or list of Device(/Reference) or Polygon Input Devices.
        B : Device(/Reference) or list of Component(/Reference) or Polygon Input Devices.
        operation : {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'} Boolean operation to perform.
        precision : float Desired precision for rounding vertex coordinates.
        num_divisions : array-like[2] of int
            number of divisions with which the geometry is divided into
            multiple rectangular regions. This allows for each region to be
            processed sequentially, which is more computationally efficient.
        max_points :
            The maximum number of vertices within the resulting polygon.
        layer : int, array-like[2], or set
            Specific layer(s) to put polygon geometry on.

    Returns:  Device
        A Component containing a polygon(s) with the boolean operations between
        the 2 input Devices performed.

    Notes
    -----
    'A+B' is equivalent to 'or'.
    'A-B' is equivalent to 'not'.
    'B-A' is equivalent to 'not' with the operands switched.
    """
    num_divisions = num_divisions or [1, 1]
    c = pg.boolean(
        A=A,
        B=B,
        operation=operation,
        precision=precision,
        num_divisions=num_divisions,
        max_points=max_points,
        layer=layer,
    )
    return gf.component_from.phidl(component=c)


if __name__ == "__main__":
    e1 = gf.components.ellipse()
    e2 = gf.components.ellipse(radii=(10, 6)).movex(2)
    e3 = gf.components.ellipse(radii=(10, 4)).movex(5)
    # gf.plot([e1, e2, e3])
    c = boolean(A=[e1, e3], B=e2, operation="A-B")
    c.show()



from typing import Any, Dict, List, Union

import pytest
from jsondiff import diff

import gdsfactory as gf
from gdsfactory.add_pins import add_settings_label
from gdsfactory.component import Component
from gdsfactory.components import factory
from gdsfactory.import_gds import add_settings_from_label, import_gds

skip_test = {
    "version_stamp",
    "extend_ports_list",
    "extend_port",
    "grating_coupler_tree",
    "compensation_path",
    "spiral_inner_io_with_gratings",
    "component_sequence",
    "straight_heater_metal_90_90",
    "straight_heater_metal_undercut_90_90",
    "mzi_phase_shifter_90_90",
}

components_to_test = set(factory.keys()) - skip_test


def tuplify(iterable: Union[List, Dict]) -> Any:
    """From a list or tuple returns a tuple."""
    if isinstance(iterable, list):
        return tuple(map(tuplify, iterable))
    if isinstance(iterable, dict):
        return {k: tuplify(v) for k, v in iterable.items()}
    return iterable


def sort_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    return {k: d[k] for k in sorted(d)}


@pytest.mark.parametrize("component_type", components_to_test)
def test_properties_components(component_type: str) -> Component:
    """Write component to GDS with setttings written on a label.
    Then import the GDS and check that the settings imported match the original settings.
    """
    cnew = gf.Component()
    c1 = factory[component_type]()
    c1ref = cnew << c1

    ignore = ("sequence", "symbol_to_component", "ports_map")
    add_settings_label(cnew, reference=c1ref, ignore=ignore)
    gdspath = cnew.write_gds_with_metadata()

    c2 = import_gds(gdspath)
    add_settings_from_label(c2)

    c1s = sort_dict(tuplify(c1.get_settings(ignore=ignore)))
    c2s = sort_dict(tuplify(c2.get_settings(ignore=ignore)))

    c1s.pop("info")
    c2s.pop("info")
    d = diff(c1s, c2s)
    # print(c1s)
    # print(c2s)
    # print(d)
    assert len(d) == 0, f"imported settings are different from original {d}"
    return c2


if __name__ == "__main__":
    # c = test_properties_components(component_type=list(component_names)[0])
    # c = test_properties_components(component_type="ring_single")
    # c = test_properties_components(component_type="mzit")
    # c = test_properties_components(component_type="bezier")
    c = test_properties_components(component_type="wire_straight")
    # c = test_properties_components(component_type="straight")
    # c = test_properties_components(component_type="grating_coupler_tree")
    # c = test_properties_components(component_type="wire")
    # c = test_properties_components(component_type="bend_circular")
    # c = test_properties_components(component_type="mzi_arm")
    # c = test_properties_components(component_type="straight_pin_passive_tapered")
    c.show()



"""
This is a sample on how to define custom components.
You can make a repo out of this file, having one custom component per file
"""
import shutil
from pathlib import Path

import numpy as np

import gdsfactory as gf
from gdsfactory.add_termination import add_gratings_and_loopback
from gdsfactory.component import Component
from gdsfactory.config import CONFIG
from gdsfactory.mask.merge_metadata import merge_metadata


def add_te(component: Component, **kwargs) -> Component:
    c = gf.routing.add_fiber_array(
        component=component,
        grating_coupler=gf.components.grating_coupler_elliptical_te,
        **kwargs,
    )
    c.test = "passive_optical_te"
    return c


def add_tm(component, **kwargs):
    c = gf.routing.add_fiber_array(
        component=component,
        grating_coupler=gf.components.grating_coupler_elliptical_tm,
        bend_radius=20,
        **kwargs,
    )
    return c


@gf.cell
def coupler_te(
    gap: float,
    length: int,
) -> Component:
    """Evanescent coupler with TE grating coupler."""
    c = gf.components.coupler(gap=gap, length=length)
    cc = add_te(c)
    return cc


@gf.cell
def spiral_te(width: float = 0.5, length: int = 2) -> Component:
    """Spiral with TE grating_coupler

    Args:
        width: waveguide width um
        lenght: cm
    """
    c = gf.c.spiral_inner_io(width=width, length=length)
    cc = add_gratings_and_loopback(
        component=c,
        grating_coupler=gf.components.grating_coupler_elliptical_te,
        bend_factory=gf.components.bend_euler,
    )
    return cc


@gf.cell
def spiral_tm(width=0.5, length=20e3):
    """Spiral with TM grating_coupler

    Args:
        width: waveguide width um
        lenght: um
    """
    c = gf.c.spiral_inner_io(width=width, length=length, waveguide_spacing=10, N=5)
    cc = add_gratings_and_loopback(
        component=c,
        grating_coupler=gf.components.grating_coupler_elliptical_tm,
        bend_factory=gf.components.bend_euler,
    )
    return cc


def test_mask(precision: float = 1e-9) -> Path:
    workspace_folder = CONFIG["samples_path"] / "mask_pack"
    build_path = workspace_folder / "build"
    mask_path = build_path / "mask"

    shutil.rmtree(build_path, ignore_errors=True)
    mask_path.mkdir(parents=True, exist_ok=True)

    gdspath = mask_path / "sample_mask.gds"
    markdown_path = gdspath.with_suffix(".md")
    json_path = gdspath.with_suffix(".json")
    test_metadata_path = gdspath.with_suffix(".tp.json")

    components = [spiral_te(length=length) for length in np.array([2, 4, 6]) * 1e4]
    components += [coupler_te(length=length, gap=0.2) for length in [10, 20, 30, 40]]
    c = gf.pack(components)
    m = c[0]
    m.name = "sample_mask"
    m.write_gds(gdspath)

    merge_metadata(gdspath=gdspath)

    assert gdspath.exists()
    assert markdown_path.exists()
    assert json_path.exists()
    assert test_metadata_path.exists()
    return gdspath


if __name__ == "__main__":
    c = test_mask()
    gf.klive.show(c)



from gdsfactory.cell import cell
from gdsfactory.component import Component


@cell
def rotate(
    component: Component,
    angle: int = 90,
) -> Component:
    """Returns rotated component inside a new component.

    Most times you just need to place a reference and rotate it.
    This rotate function just encapsulates the rotated reference into a new component.

    Args:
        component:
        angle: in degrees
    """
    component_new = Component()
    component_new.component = component
    ref = component_new.add_ref(component)
    ref.rotate(angle)
    component_new.add_ports(ref.ports)
    return component_new


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.mmi1x2()
    cr = rotate(component=c)
    cr.show()
    # print(component_rotated)

    # component_rotated.pprint
    # component_netlist = component.get_netlist()
    # component.pprint_netlist()



from gdsfactory.component import Component, ComponentReference
from gdsfactory.hash_points import hash_points
from gdsfactory.port import Port


def get_route_electrical_shortest_path(port1: Port, port2: Port) -> ComponentReference:
    """Returns polygon reference that connects two ports
    with a polygon that takes the shortest path

    Args:
        port1: source port
        port2: destination port
    """
    points = [port1.midpoint, port2.midpoint]
    name = f"zz_conn_{hash_points(points)}"
    c = Component(name=name)
    layer = port1.layer
    p1x0 = port1.endpoints[0][0]
    p1y0 = port1.endpoints[0][1]
    p1x1 = port1.endpoints[1][0]
    p1y1 = port1.endpoints[1][1]

    x = [p1x0, p1x1]
    y = [p1y0, p1y1]
    p1y0 = min(y)
    p1y1 = max(y)
    p1x0 = min(x)
    p1x1 = max(x)

    p2x0 = port2.endpoints[0][0]
    p2y0 = port2.endpoints[0][1]
    p2x1 = port2.endpoints[1][0]
    p2y1 = port2.endpoints[1][1]

    x = [p2x0, p2x1]
    y = [p2y0, p2y1]
    p2y0 = min(y)
    p2y1 = max(y)
    p2x0 = min(x)
    p2x1 = max(x)

    if port1.orientation in [90, 270]:
        c.add_polygon(
            ([(p1x0, p1y0), (p1x1, p1y0), (p2x1, p2y1), (p2x0, p2y0)]), layer=layer
        )
    elif port1.orientation == 180:
        c.add_polygon(
            ([(p1x0, p1y1), (p1x1, p1y0), (p2x1, p2y0), (p2x0, p2y1)]), layer=layer
        )
    elif port1.orientation == 0:
        c.add_polygon(
            ([(p1x0, p1y0), (p1x1, p1y1), (p2x1, p2y0), (p2x0, p2y1)]), layer=layer
        )

    return c.ref()


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.components.pad import pad_array

    c = Component("mzi_with_pads")
    mzi = gf.components.mzi_phase_shifter()
    pads = pad_array(columns=3, orientation=270)
    p = c << pads
    c << mzi
    p.move((-150, 200))
    ports_pads = list(p.ports.values())
    ports_mzi = mzi.get_ports_list()

    for p1, p2 in zip(ports_pads, ports_mzi):
        c.add(get_route_electrical_shortest_path(p1, p2))
    c.show()



from typing import Dict, List, Tuple

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.types import Layer

DIRECTION_TO_ANGLE = {"W": 180, "E": 0, "N": 90, "S": 270}


@cell
def rectangle(
    size: Tuple[float, float] = (4.0, 2.0),
    layer: Layer = gf.LAYER.WG,
    centered: bool = False,
    ports: Dict[str, List[Tuple[float, float, float]]] = None,
    **port_settings
) -> Component:
    """rectangle

    Args:
        size: (tuple) Width and height of rectangle.
        layer: (int, array-like[2], or set) Specific layer(s) to put polygon geometry on.
        centered: True sets center to (0, 0)
        ports: {direction: [(x, y, width), ...]} direction: 'W', 'E', 'N' or 'S'

    """
    c = gf.Component()
    w, h = size

    if centered:
        points = [
            [-w / 2.0, -h / 2.0],
            [-w / 2.0, h / 2],
            [w / 2, h / 2],
            [w / 2, -h / 2.0],
        ]
    else:
        points = [[w, h], [w, 0], [0, 0], [0, h]]
    c.add_polygon(points, layer=layer)

    i = 0
    if ports:
        for direction, list_port_params in ports.items():
            assert direction in "NESW"
            angle = DIRECTION_TO_ANGLE[direction]
            for x, y, width in list_port_params:
                c.add_port(
                    name="{}".format(i),
                    orientation=angle,
                    midpoint=(x, y),
                    width=width,
                    layer=layer,
                    **port_settings
                )
                i += 1

    gf.port.auto_rename_ports(c)
    return c


if __name__ == "__main__":
    c = rectangle(size=("hi", 2), ports={"N": [(0, 1, 4)]}, centered=True, layer=(2, 3))
    print(c.ports)
    print(c.name)
    c.show()



"""One problem is that when we add references we have to make sure they have unique names.

The photonics package `gf. has a cell decorator that names the objects that it produces depending on the parameters that we pass them
"""


from typing import Union

import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def straight_cell(width: Union[float, int] = 10, height: int = 1) -> Component:
    """Returns straight with automatic name."""
    wg = gf.Component("straight")
    wg.add_polygon([(0, 0), (width, 0), (width, height), (0, height)])
    wg.add_port(name="wgport1", midpoint=[0, height / 2], width=height, orientation=180)
    wg.add_port(
        name="wgport2", midpoint=[width, height / 2], width=height, orientation=0
    )
    return wg


def test_autoname() -> None:
    c = straight_cell()
    assert c.name == "straight_cell"

    c = straight_cell(width=0.5)
    assert c.name == "straight_cell_width500n"


if __name__ == "__main__":
    c = straight_cell()
    print(c)

    c = straight_cell(width=0.5)
    print(c)



import pathlib
from pathlib import Path
from typing import Dict, Tuple

from gdsfactory.component import Component
from gdsfactory.config import CONFIG
from gdsfactory.name import dict2name
from gdsfactory.tech import LAYER


def get_sparameters_path(
    component: Component,
    layer_to_material: Dict[Tuple[int, int], str],
    layer_to_thickness_nm: Dict[Tuple[int, int], int],
    dirpath: Path = CONFIG["sp"],
) -> Path:
    """Returns Sparameters filepath.

    Args:
        component:
        dirpath
        layer_to_material: GDSlayer to material alias (see aliases in gf.sp.write)
        layer_to_thickness_nm: GDSlayer to thickness (nm)
    """
    dirpath = pathlib.Path(dirpath)
    dirpath = (
        dirpath / component.function_name
        if hasattr(component, "function_name")
        else dirpath
    )
    dirpath.mkdir(exist_ok=True, parents=True)
    material2nm = {
        layer_to_material[layer]: layer_to_thickness_nm[layer]
        for layer in layer_to_thickness_nm.keys()
        if tuple(layer) in component.get_layers()
    }
    suffix = dict2name(**material2nm)
    return dirpath / f"{component.get_name_long()}_{suffix}.dat"


def test_get_sparameters_path() -> None:
    import gdsfactory as gf

    layer_to_thickness_nm_sample = {
        LAYER.WG: 220,
        LAYER.SLAB90: 90,
    }
    layer_to_material_sample = {
        LAYER.WG: "si",
        LAYER.SLAB90: "si",
    }

    c = gf.components.straight()
    p = get_sparameters_path(
        component=c,
        layer_to_thickness_nm=layer_to_thickness_nm_sample,
        layer_to_material=layer_to_material_sample,
    )
    assert p.stem == "straight_si220", p.stem

    c = gf.components.straight(layer=LAYER.SLAB90)
    p = get_sparameters_path(
        c,
        layer_to_thickness_nm=layer_to_thickness_nm_sample,
        layer_to_material=layer_to_material_sample,
    )
    assert p.stem == "straight_layer3_0_si90", p.stem


if __name__ == "__main__":
    # import gdsfactory as gf
    # c = gf.components.straight()
    # p = get_sparameters_path(c)
    # print(p)

    test_get_sparameters_path()



import re
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.sp.get_sparameters_path import get_sparameters_path
from gdsfactory.tech import LAYER_STACK, LayerStack


def get_ports(line: str) -> Tuple[str, str]:
    """Returns 2 port labels strings from interconnect file."""
    line = line.replace('"', "")
    line = line.replace("(", "")
    line_fields = line.split(",")
    port1 = line_fields[0]
    port2 = line_fields[3]
    return port1, port2


def read_sparameters_lumerical(
    filepath, numports: int
) -> Tuple[Tuple[str, ...], np.array, np.ndarray]:
    r"""Returns Sparameters from Lumerical interconnect export file.

    Args:
        filepath: Sparameters filepath (interconnect format)
        numports: number of ports

    Returns:
        port_names: list of port labels
        F: frequency 1d np.array
        S: Sparameters np.ndarray matrix

    """
    F = []
    S = []
    port_names = []

    with open(filepath, "r") as fid:
        for _i in range(numports):
            port_line = fid.readline()
            m = re.search(r'\[".*",', port_line)
            if m:
                port = m.group(0)
                port_names.append(port[2:-2])
        line = fid.readline()
        port1, port2 = get_ports(line)
        line = fid.readline()
        numrows = int(tuple(line[1:-2].split(","))[0])
        S = np.zeros((numrows, numports, numports), dtype="complex128")
        r = m = n = 0
        for line in fid:
            if line[0] == "(":
                if "transmission" in line:
                    port1, port2 = get_ports(line)
                continue
            data = line.split()
            data = list(map(float, data))
            if m == 0 and n == 0:
                F.append(data[0])

            i = port_names.index(port1)
            j = port_names.index(port2)
            S[r, i, j] = data[1] * np.exp(1j * data[2])
            r += 1
            if r == numrows:
                r = 0
                m += 1
                if m == numports:
                    m = 0
                    n += 1
                    if n == numports:
                        break

    # port_names.reverse()
    # print(len(F), S.shape, len(port_names))
    return (tuple(port_names), np.array(F), S)


def test_read_sparameters_2port_bend():
    filepath = gf.CONFIG["sp"] / "bend_circular" / "bend_circular_S220.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=2)
    print(port_names)
    assert port_names == ("N0", "W0")


def test_read_sparameters_2port_straight():
    filepath = gf.CONFIG["sp"] / "straight" / "straight_S220.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=2)
    print(port_names)
    assert len(f) == 500
    assert port_names == ("E0", "W0")


def test_read_sparameters_3port_mmi1x2():
    filepath = gf.CONFIG["sp"] / "mmi1x2" / "mmi1x2_S220.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=3)
    print(port_names)
    assert len(f) == 500
    assert port_names == ("E0", "E1", "W0")


def test_read_sparameters_4port_mmi2x2():
    filepath = gf.CONFIG["sp"] / "mmi2x2" / "mmi2x2_S220.dat"
    port_names, f, s = read_sparameters_lumerical(filepath=filepath, numports=4)
    print(port_names)
    assert len(f) == 500
    assert port_names == ("E0", "E1", "W0", "W1")


def read_sparameters_component(
    component: Component,
    layer_stack: LayerStack = LAYER_STACK,
    layer_to_material: Optional[Dict[Tuple[int, int], str]] = None,
    layer_to_thickness_nm: Optional[Dict[Tuple[int, int], int]] = None,
    dirpath: Path = gf.CONFIG["sp"],
) -> Tuple[List[str], np.array, np.ndarray]:
    r"""Returns Sparameters from Lumerical interconnect export file.

    Args:
        component: Component
        dirpath: path where to look for the Sparameters
        layer_to_material: layer to material dict
        layer_to_thickness_nm: layer to thickness (nm)

    Returns:
        port_names: list of port labels
        F: frequency 1d np.array
        S: Sparameters np.ndarray matrix


    the Sparameters file have Lumerical format
    https://support.lumerical.com/hc/en-us/articles/360036107914-Optical-N-Port-S-Parameter-SPAR-INTERCONNECT-Element#toc_5
    """

    assert isinstance(component, gf.Component)
    filepath = get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_to_material=layer_to_material or layer_stack.get_layer_to_material(),
        layer_to_thickness_nm=layer_to_thickness_nm
        or layer_stack.get_layer_to_thickness_nm(),
    )
    numports = len(component.ports)
    assert (
        filepath.exists()
    ), f"Sparameters for {component.name} not found in {filepath}"
    assert numports > 1, f"number of ports = {numports} and needs to be > 1"
    return read_sparameters_lumerical(filepath=filepath, numports=numports)


def read_sparameters_pandas(
    component: Component,
    layer_to_material: Optional[Dict[Tuple[int, int], str]] = None,
    layer_to_thickness_nm: Optional[Dict[Tuple[int, int], int]] = None,
    dirpath: Path = gf.CONFIG["sp"],
) -> pd.DataFrame:
    filepath = get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_to_material=layer_to_material or LAYER_STACK.get_layer_to_material(),
        layer_to_thickness_nm=layer_to_thickness_nm
        or LAYER_STACK.get_layer_to_thickness_nm(),
    )
    df = pd.read_csv(filepath.with_suffix(".csv"))
    df.index = df["wavelength_nm"]
    return df


if __name__ == "__main__":
    # test_read_sparameters_2port_straight()
    # test_read_sparameters_2port_bend()
    # test_read_sparameters_3port_mmi1x2()
    # test_read_sparameters_4port_mmi2x2()
    s = read_sparameters_component(gf.components.mmi2x2())
    # print(s[0])
    # print(s)
